# Import functions from other modules ============================
# from io_funs import LoadSave

# Import python libraries ========================================

# ******* Standard Data Manipulation / Statistical Libraries *****
import pandas as pd
pd.options.mode.chained_assignment = None  # Suppress warnings
import numpy as np
import pickle as pk

from typing import Union


# ******* Data Visualization Libraries ****************************

import matplotlib.pyplot as plt

import seaborn as sns

from bokeh.io import output_notebook
output_notebook()
from bokeh.plotting import show,figure
TOOLTIPS = [
    ("index", "$index"),
    ("(x,y)", "($x, $y)"),
]

# ******** Data science / Machine learning Libraries ***************
import tensorflow as tf
from tensorflow.keras.layers import Conv1D, MaxPooling1D, Flatten, Dense, Dropout
from tensorflow.keras.callbacks import EarlyStopping

tf.get_logger().setLevel('ERROR')

from tensorflow import keras
from tensorflow.keras.models import save_model


class TrainRegressorCNN:
    """
    Train Convolutional Neural Networks model using regression approach

    Parameters
    -----------
    X1_train : array
        Row-StandardScaled input spectra for training.
    X1_val : array
        Row-StandardScaled input spectra for validation.
    X1_test : array
        Row-StandardScaled input spectra for testing.

    X2_train : array
        Col-StandardScaled Mix Max of all rows of input spectra for training.
    X2_val : array
        Col-StandardScaled Mix Max of all rows of input spectra for validation.
    X2_test : array
        Col-StandardScaled Mix Max of all rows of input spectra for testing.

    y1_train : array
        Col-StandardScaled target feature 1 for training.
    y1_val : array
        Col-StandardScaled target feature 1 for validation.
    y1_test : array
        Col-StandardScaled target feature 1 for testing.

    y2_train : array
        Col-StandardScaled target feature 2 for training.
    y2_val : array
        Col-StandardScaled target feature 2 for validation.
    y2_test : array
        Col-StandardScaled target feature 2 for testing.

    y3_train : array
        Col-StandardScaled target feature 3 for training.
    y3_val : array
        Col-StandardScaled target feature 3 for validation.
    y3_test : array
        Col-StandardScaled target feature 3 for testing.

    y4_train : array
        Col-StandardScaled target feature 4 for training.
    y4_val : array
        Col-StandardScaled target feature 4 for validation.
    y4_test : array
        Col-StandardScaled target feature 4 for testing.
    """
    def __init__(self,
                 X1_train: Union[np.ndarray, list],
                 X1_val: Union[np.ndarray, list],
                 X1_test: Union[np.ndarray, list],
                 X2_train: Union[np.ndarray, list],
                 X2_val: Union[np.ndarray, list],
                 X2_test: Union[np.ndarray, list],
                 y1_train: Union[np.ndarray, list],
                 y1_val: Union[np.ndarray, list],
                 y1_test: Union[np.ndarray, list],
                 y2_train: Union[np.ndarray, list],
                 y2_val: Union[np.ndarray, list],
                 y2_test: Union[np.ndarray, list],
                 y3_train: Union[np.ndarray, list],
                 y3_val: Union[np.ndarray, list],
                 y3_test: Union[np.ndarray, list],
                 y4_train: Union[np.ndarray, list],
                 y4_val: Union[np.ndarray, list],
                 y4_test: Union[np.ndarray, list]
                 ):

        # train, val, test sets for main features (104 wavelengths)
        self.X1_train, self.X1_val, self.X1_test = X1_train, X1_val, X1_test

        # train, val, test sets for input 2 (Min and Max 2 features)
        self.X2_train, self.X2_val, self.X2_test = X2_train, X2_val, X2_test

        # train, val, test sets for target features
        self.y1_train, self.y1_val, self.y1_test = y1_train, y1_val, y1_test
        self.y2_train, self.y2_val, self.y2_test = y2_train, y2_val, y2_test
        self.y3_train, self.y3_val, self.y3_test = y3_train, y3_val, y3_test
        self.y4_train, self.y4_val, self.y4_test = y4_train, y4_val, y4_test

    def build_model(self,
                    hyperparameters, # dic
                    ):
        """
        Build a CNN model with the given hyperparameters.

        Parameters
        ----------
        hyperparameters : dict
            A dictionary containing hyperparameter settings.

            hyperparameters keys includes:
                - 'Conv__num_blocks' (int): Number of blocks in the CNN model.
                - 'Conv__num_layers_per_block' (int): Number of layers in each convolutional block.
                - 'Conv__num_filters' (list): Number of filters in each layer.
                - 'Conv__kernel_size' (int): Size of the convolutional kernel.
                - 'Conv__MaxPooling1D' (bool): MaxPooling1D size.

                - 'FC1__num_blocks' (int): Number of blocks in the first fully connected (FC1) part.
                - 'FC1_num_layers_per_block' (int): Number of layers in each FC1 block.
                - 'FC1__units' (list): Number of units in each FC1 layer.
                - 'FC1__dropout' (float): Dropout rate for FC1 layers.

                - 'FC2__num_blocks' (int): Number of blocks in the second fully connected (FC2) part.
                - 'FC2_num_layers_per_block' (int): Number of layers in each FC2 block.
                - 'FC2__units' (list): Number of units in each FC2 layer.
                - 'FC2__dropout' (float): Dropout rate for FC2 layers.

                - 'learning_rate' (float): Learning rate for the model.

        Example
        --------
        >>> hyperparameters = {
        >>>      'Conv__MaxPooling1D': 2,
        >>>      'Conv__num_blocks': 1,
        >>>      'Conv__num_layers_per_block': 3,
        >>>      'Conv__num_filters': 4,
        >>>      'Conv__kernel_size': 6,
        >>>      'FC__NumberLayers': 4,
        >>>      'FC1__num_blocks' : 1,
        >>>      'FC1_num_layers_per_block': 4,
        >>>      'FC1__dropout': 0.09889223768186726,
        >>>      'FC1__units': 128,
        >>>      'FC2__num_blocks' : 1,
        >>>      'FC2_num_layers_per_block':2,
        >>>      'FC2__dropout': 0.0024609140719442646,
        >>>      'FC2__units': 64,
        >>>      'learning_rate': 4.9946842008422193e-05}

        Returns
        -------
        object
            Pre-build tf.keras.Model CNN model.

        """

        Conv__num_blocks = hyperparameters['Conv__num_blocks']
        Conv__num_layers_per_block = hyperparameters['Conv__num_layers_per_block']
        Conv__num_filters = hyperparameters['Conv__num_filters']
        Conv__kernel_size = hyperparameters['Conv__kernel_size']
        Conv__MaxPooling1D = hyperparameters['Conv__MaxPooling1D']

        FC1__num_blocks = hyperparameters['FC1__num_blocks']
        FC1_num_layers_per_block = hyperparameters['FC1_num_layers_per_block']
        FC1__units = hyperparameters['FC1__units']
        FC1__dropout = hyperparameters['FC1__dropout']

        FC2__num_blocks = hyperparameters['FC2__num_blocks']
        FC2_num_layers_per_block = hyperparameters['FC2_num_layers_per_block']
        FC2__units = hyperparameters['FC2__units']
        FC2__dropout = hyperparameters['FC2__dropout']

        self.learning_rate = hyperparameters['learning_rate']

        # Define the input layer
        input_layer_1 = tf.keras.layers.Input(shape=(104, 1))
        input_layer_2 = tf.keras.layers.Input(shape=(2,))

        # Start building the model using the input layer
        x = input_layer_1

        # Build the specified number of blocks using for loops
        for block in range(Conv__num_blocks):
            # Build the specified number of layers in each block using for loops
            for layer in range(Conv__num_layers_per_block):
                x = Conv1D(filters=Conv__num_filters * ((block + layer + 1) * 2) ** 2,
                           kernel_size=Conv__kernel_size,
                           activation='relu',
                           padding='same',
                           kernel_initializer='he_normal',
                           name='Conv__B' + str(block + 1) + '_L' + str(layer + 1)
                           )(x)

            # Add a MaxPooling layer at the end of each block
            x = MaxPooling1D(pool_size=(Conv__MaxPooling1D),
                             name='MaxPool1D__B' + str(block + 1) + '_L' + str(layer + 1))(x)

        # Flatten the output of the last block
        x = Flatten()(x)

        for block in range(FC1__num_blocks):
            for layer in range(FC1_num_layers_per_block):
                x = Dense(FC1__units * ((block + layer + 1) * 2) ** 2,
                          activation='relu',
                          name='FC1__B' + str(block + 1) + '_L' + str(layer + 1))(x)

            x = Dropout(FC1__dropout,
                        name='FC1__Dropout__B' + str(block + 1) + '_L' + str(layer + 1))(x)

        # Concatenate the outputs from the convolutional layers and dense layer
        x = tf.keras.layers.concatenate([x, input_layer_2],
                                        name='Concatenated_Layer')

        # Add a dense layer for classification
        for block in range(FC2__num_blocks):
            for layer in range(FC2_num_layers_per_block):
                x = Dense(FC2__units * ((block + layer + 1) * 2) ** 2,
                          activation='relu',
                          name='FC2__B' + str(block + 1) + '_L' + str(layer + 1))(x)

            x = Dropout(FC2__dropout,
                        name='FC2__Dropout__B' + str(block + 1) + '_L' + str(layer + 1))(x)

        ######### 3rd FC Block: gravity  ##############################

        out__gravity = Dense(1,
                             activation='linear',
                             # kernel_initializer = 'he_normal',
                             name='gravity')(x)

        ######### 3rd FC Block: c_o_ratio  ##############################
        out__c_o_ratio = Dense(1,
                               activation='linear',
                               # kernel_initializer = 'he_normal',
                               name='c_o_ratio')(x)

        ######### 3rd FC Block: metallicity  ##############################
        out__metallicity = Dense(1,
                                 activation='linear',
                                 # kernel_initializer = 'he_normal',
                                 name='metallicity')(x)

        ######### 3rd FC Block: temperature  ##############################
        out__temperature = Dense(1,
                                 activation='linear',
                                 name='temperature')(x)

        ######### OUTPUT   ################################################
        # Create the model with two inputs and two outputs
        model = tf.keras.Model(inputs=[input_layer_1, input_layer_2],
                               outputs=[out__gravity, out__c_o_ratio, out__metallicity, out__temperature])

        self.model = model

        print(model.summary())

    def fit_cnn_model(self,
                      batch_size = 32,
                      budget=3):
        """
        Fit the pre-build CNN model

        Returns:
            Training history (Loss values for train)
            Trained model
        """
        model = self.model
        # Compile the model with an optimizer, loss function, and metrics
        model.compile(loss='huber_loss',
                           optimizer=keras.optimizers.Adam(learning_rate=self.learning_rate),
                           metrics=['mae'])

        early_stop = EarlyStopping(monitor='loss', min_delta=4e-4, patience=50, mode='auto', \
                                   restore_best_weights=True)

        # YOU CAN ADD FUNCTION HERE TO ADD NOISE
        history = self.model.fit(x=[self.X1_train, self.X2_train],
                                 y=[self.y1_train, self.y2_train, self.y3_train, self.y4_train],
                                 # self.x_train, self.y_train,
                                 batch_size= batch_size,  # config['batch_size'], # self.batch_size,
                                 validation_data=(
                                 [self.X1_val, self.X2_val], [self.y1_val, self.y2_val, self.y3_val, self.y4_val]),
                                 # validation_split=0.2,
                                 epochs=int(budget),
                                 verbose=1,
                                 callbacks=[early_stop],
                                 )
        self.model = model
        self.history = history
        return history, model
#         train_score = model.evaluate(x = [self.X1_train, self.X2_train],
#                                      y = [self.y1_train, self.y2_train, self.y3_train, self.y4_train],
#                                      verbose=0)
#         val_score   = model.evaluate(x = [self.X1_val, self.X2_val],
#                                      y = [self.y1_val, self.y2_val, self.y3_val, self.y4_val],
#                                      verbose=0)
#         test_score  = model.evaluate(x = [self.X1_test, self.X2_test],
#                                      y = [self.y1_test, self.y2_test, self.y3_test, self.y4_test],
#                                      verbose=0)


import os
import pickle as pk
from tensorflow.keras.models import load_model

# from tensorflow.keras.models import model_from_json

__reference_data__ = os.getenv("TelescopeML_reference_data")
# print(__reference_data__)



class LoadSave:
    """
    Load and Save trained operators, models, and datasets
    """

    def __init__(self,
                 ml_model_str,
                 ml_method,
                 is_feature_improved = None,
                 is_augmented = None,
                 is_tuned = None,
                 ):
        self.ml_model_str = ml_model_str
        self.ml_method = ml_method
        self.is_feature_improved = is_feature_improved
        self.is_augmented = is_augmented
        self.is_tuned = is_tuned
        self.base_path = os.path.join(__reference_data__,'trained_ML_models')
        # '../outputs/trained_models/'

    def create_generic_path(self, output_indicator):
        """
        Create the generic path for saving or loading the trained model

        Inputs:
        -------
        - indicator (str): Indicator for the model type

        Returns:
        --------
        - generic_path (str): The generic path for saving or loading the model
        """
        file_name = f'{output_indicator}__{self.ml_model_str}__' \
                    f'{self.ml_method}'

        generic_path = os.path.join(self.base_path, file_name)
        return generic_path

    def load_or_dump_trained_object(self,
                                    trained_object,
                                    output_indicator,
                                    load_or_dump='dump'):
        """
        Load or save the trained object

        Inputs:
        -------
        - trained_object : The object to be saved or loaded
        - indicator (str): Indicator for the type of trained object
        - load_or_dump (str): 'dump' or 'load'
        """
        generic_path = self.create_generic_path(output_indicator)

        if load_or_dump == 'dump':
            with open(generic_path, 'wb') as file:
                pk.dump(trained_object, file)
        elif load_or_dump == 'load':
            with open(generic_path, 'rb') as file:
                return pk.load(file)


def load_or_dump_trained_model_CNN(
                                   trained_model = None,
                                   output_indicator='TrainedCNN',
                                   load_or_dump='dump'):

    path_architecture = os.path.join(__reference_data__,
                         'trained_ML_models/trained_CNN_architecture_'+output_indicator+'.h5',
                         )

    path_history = os.path.join(__reference_data__,
                         'trained_ML_models/trained_CNN_history_'+output_indicator+'.pkl',
                         )

    path_weights = os.path.join(__reference_data__,
                         'trained_ML_models/trained_CNN_weights_'+output_indicator+'.h5',
                         )

    """
    Load or save the trained CNN model

    Inputs:
    -------
    - trained_model : The trained CNN model
    - indicator (str): Indicator for the type of trained model
    - load_or_dump (str): 'dump' or 'load'
    """
    # json_path = self.create_generic_path(f'{indicator}_json')
    # weights_path = self.create_generic_path(f'{indicator}_weights')

    if load_or_dump == 'dump':
        trained_model.trained_model.save(path_architecture)
        trained_model.trained_model.save_weights(path_weights)

        with open(path_history, 'wb') as file:
            pk.dump(trained_model.history.history, file)

    elif load_or_dump == 'load':
        loaded_model = load_model(path_architecture)
        loaded_model.load_weights(path_weights)

        # Loading the saved history object
        with open(path_history, 'rb') as file:
                history = pk.load(file)

        return loaded_model, history




# Import functions from other modules ============================
from .IO_utils import *

# Import python libraries ========================================

# ******* Standard Data Manipulation / Statistical Libraries *****
import pandas as pd

pd.options.mode.chained_assignment = None  # Suppress warnings
import numpy as np
import pickle as pk

from typing import List, Union, Dict
from sklearn.base import BaseEstimator

# ******* Data Visualization Libraries ****************************

import matplotlib.pyplot as plt

from bokeh.io import output_notebook

output_notebook()
from bokeh.plotting import show, figure

TOOLTIPS = [
    ("index", "$index"),
    ("(x,y)", "($x, $y)"),
]

# ******** Data science / Machine learning Libraries ***************
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler

import tensorflow as tf

tf.get_logger().setLevel('ERROR')


# ===============================================================================
# ==================                                           ==================
# ==================            Train CNN Regression           ==================
# ==================                                           ==================
# ===============================================================================

class DataProcessor:
    """
    Perform various tasks to process the datasets, including:

    - Prepare inputs and outputs
    - Split the dataset into training, validation, and test sets
    - Scale/normalize the data
    - Visualize the data
    - Conduct feature engineering

    Parameters
    ----------
    feature_values : np.ndarray
        Flux arrays (input data).
    feature_names : List[str]
        Name of wavelength in micron.
    target_values : np.ndarray
        Target variable array (e.g., Temperature, Gravity, Carbon_to_Oxygen, Metallicity).
    target_name : str
        Name of the target variable.
    is_hyperparam_tuned : str
        Indicates whether hyperparameters are tuned or not ('yes' or 'no').
    param_grid : dict, optional
        ML hyperparameters to be tuned (used if is_hyperparam_tuned = 'yes').
    spectral_resolution : int, optional
        Resolution of the synthetic spectra used to generate the dataset.
    feature_improvement_method : str
        Method used for feature improvement ('no', 'pca', 'RFE').
    augmentation_method : str
        Indicates if an augmented dataset is used ('no' or method name).
    ml_model : BaseEstimator, optional
        ML model object from sklearn package.
    ml_model_str : str, optional
        Name of the ML model.
    ml_method : str, optional
        Machine learning method ('regression' or 'classification').

    """
    def __init__(
            self,
            feature_values: Union[np.ndarray] = None,
            feature_names: Union[List[str]] = None,
            target_values: Union[np.ndarray] = None,
            target_name: Union[str] = None,
            is_tuned: str = 'no',
            param_grid: Union[None, Dict] = None,
            spectral_resolution: Union[None, int] = None,
            is_feature_improved: str = 'no',
            is_augmented: str = 'no',
            ml_model: Union[None, BaseEstimator] = None,
            ml_model_str: Union[None, str] = None,
            ml_method: str = 'regression',
    ):

        self.feature_values = feature_values
        self.feature_names = feature_names
        self.target_values = target_values
        self.target_name = target_name
        self.is_tuned = is_tuned
        self.param_grid = param_grid
        self.spectral_resolution = spectral_resolution
        self.is_feature_improved = is_feature_improved
        self.is_augmented = is_augmented
        self.ml_model = ml_model
        self.ml_model_str = ml_model_str
        self.ml_method = ml_method
        self.LoadSave = LoadSave(ml_model_str,
                                 ml_method,
                                 is_feature_improved,
                                 is_augmented,
                                 is_tuned)

    def split_train_test(self, test_size=0.1):
        """
        Split the loaded set into train and test sets

        Parameters
        ----------
        test_size : float
            The proportion of the dataset to include in the test split.

        Returns
        -------
        self.X_train : array
            X train set.
        self.X_test : array
            X test set.
        self.y_train : array
            y train set.
        self.y_test : array
            y test set.

        References
        ----------
        link: `sklearn.model_selection.train_test_split <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html>`_

        """
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.feature_values,
                                                                                self.target_values,
                                                                                test_size=test_size,
                                                                                shuffle=True,
                                                                                random_state=42)

    def split_train_validation_test(self,
                                    test_size=0.1,
                                    val_size=0.1,
                                    random_state_=42):
        """
        Split the loaded dataset into train, validation, and test sets.

        Parameters
        ----------
        test_size : float
            Proportion of the dataset to include in the test split.
        val_size : float
            Proportion of the remaining train dataset to include in the validation split.

        Returns
        -------
        self.X_train : array
            Used to train the machine learning model.
        self.X_val : array
            Used to validate the machine learning model.
        self.X_test : array
            Used to evaluate the machine learning model.
        self.y_train : array
            Targets used for training the models.
        self.y_val : array
            Targets used for validating the models.
        self.y_test : array
            Targets used for testing the models.

        References
        ----------
        link: `sklearn.model_selection.train_test_split <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html>`_
        """
        X_train, X_test, y_train, y_test = train_test_split(
            self.feature_values,
            self.target_values,
            test_size=test_size,
            shuffle=True,
            random_state=random_state_
        )
        self.X_test, self.y_test = X_test, y_test

        self.X_train, self.X_val, self.y_train, self.y_val = train_test_split(
            X_train,
            y_train,
            test_size=val_size,
            shuffle=True,
            random_state=random_state_
        )

    def normalize_X_column_wise(self,
                                output_indicator='Trained_Normalizer_X_ColWise',
                                X_train=None,
                                X_val=None,
                                X_test=None,
                                print_model=False):
        """
        Normalize features/column variables to a specified range [0, 1].

        Parameters
        ----------
        X_train : array
            Training feature matrix.
        X_val : array
            Validation feature matrix.
        X_test : array
            Test feature matrix.
        print_model : bool, optional
            Whether to print the trained normalizer model.

        Returns
        -------
        self.X_train_normalized_columnwise : array
            Normalized training feature matrix.
        self.X_val_normalized_columnwise : array
            Normalized validation feature matrix.
        self.X_test_normalized_columnwise : array
            Normalized test feature matrix.
        self.normalize_X_ColumnWise : MinMaxScaler
            Trained normalizer object.

        """
        # Set default values if None is provided
        X_train = self.X_train if X_train is None else X_train
        X_val = self.X_val if X_val is None else X_val
        X_test = self.X_test if X_test is None else X_test

        normalizer = MinMaxScaler(feature_range=(0, 1))
        self.X_train_normalized_columnwise = normalizer.fit_transform(X_train)
        self.X_val_normalized_columnwise = normalizer.transform(X_val)
        self.X_test_normalized_columnwise = normalizer.transform(X_test)
        self.normalize_X_ColumnWise = normalizer

        self.LoadSave.load_or_dump_trained_object(
            trained_object=self.normalize_X_ColumnWise,
            output_indicator=output_indicator,
            load_or_dump='dump')

        if print_model:
            print(normalizer)

    def normalize_X_row_wise(self,
                             output_indicator='Trained_Normalizer_X_RowWise',
                             X_train=None,
                             X_val=None,
                             X_test=None,
                             print_model=False):
        """
        Normalize observations/instances/row variables to a specified range [0, 1].

        Transform your data such that its values are within the specified range [0, 1].

        Parameters
        ----------
        X_train : array
            Training feature matrix.
        X_val : array
            Validation feature matrix.
        X_test : array
            Test feature matrix.
        print_model : bool, optional
            Whether to print the trained normalizer model.

        Returns
        -------
        self.X_train_normalized_rowwise : array
            Normalized training feature matrix.
        self.X_val_normalized_rowwise : array
            Normalized validation feature matrix.
        self.X_test_normalized_rowwise : array
            Normalized test feature matrix.
        """
        # Set default values if None is provided
        X_train = self.X_train if X_train is None else X_train
        X_val = self.X_val if X_val is None else X_val
        X_test = self.X_test if X_test is None else X_test

        normalizer = MinMaxScaler(feature_range=(0, 1))
        self.X_train_normalized_rowwise = normalizer.fit_transform(X_train.T).T
        self.X_val_normalized_rowwise = normalizer.fit_transform(X_val.T).T
        self.X_test_normalized_rowwise = normalizer.fit_transform(X_test.T).T
        self.normalize_X_RowWise = normalizer

        self.LoadSave.load_or_dump_trained_object(
            trained_object=self.normalize_X_RowWise,
            output_indicator=output_indicator,
            load_or_dump='dump')

        if print_model:
            print(normalizer)

    def normalize_y_column_wise(self,
                                output_indicator =  'Trained_Normalizer_y_ColWise',
                                y_train=None,
                                y_val=None, y_test=None, print_model=False):
        """
        Scale target variable (y) and column-wise feature variables to a specified range.
        Transform the data such that its values are within the specified range [0, 1].

        Parameters
        -----------
        y_train : array
            Training target variable array.
        y_val : array
            Validation target variable array.
        y_test : array
            Test target variable array.
        print_model : bool
            Whether to print the trained scaler model.

        Returns
        -------
        self.y_train_normalized_columnwise : array
            Scaled training target variable array.
        self.y_val_normalized_columnwise : array
            Scaled validation target variable array.
        self.y_test_normalized_columnwise : array
            Scaled test target variable array.
        """
        # Set default values if None is provided
        y_train = self.y_train if y_train is None else y_train
        y_val = self.y_val if y_val is None else y_val
        y_test = self.y_test if y_test is None else y_test

        scaler_y = MinMaxScaler(feature_range=(0, 1))
        self.y_train_normalized_columnwise = scaler_y.fit_transform(y_train)
        self.y_val_normalized_columnwise = scaler_y.transform(y_val)
        self.y_test_normalized_columnwise = scaler_y.transform(y_test)
        self.normalize_y_ColumnWise = scaler_y

        self.LoadSave.load_or_dump_trained_object(
            trained_object=self.normalize_y_ColumnWise,
            output_indicator=output_indicator,
            load_or_dump='dump')

        if print_model:
            print(scaler_y)

    def standardize_y_column_wise(self,
                                  output_indicator = 'Trained_StandardScaler_y_ColWise',
                                  y_train=None,
                                  y_val=None,
                                  y_test=None,
                                  print_model=False):
        """
        Standardize target variable (y) column-wise by removing the mean and scaling to unit variance.
        Transform the data such that its distribution will have a mean value of 0 and a standard deviation of 1.

        Parameters
        -----------
        y_train : array
            Training target variable array.
        y_val : array
            Validation target variable array.
        y_test : array
            Test target variable array.
        print_model : bool
            Whether to print the trained scaler model.

        Returns
        -------
        self.y_train_standardized_columnwise : array
            Standardized training target variable array.
        self.y_val_standardized_columnwise : array
            Standardized validation target variable array.
        self.y_test_standardized_columnwise : array
            Standardized test target variable array.
        """

        # Set default values if None is provided
        y_train = self.y_train if y_train is None else y_train
        y_val = self.y_val if y_val is None else y_val
        y_test = self.y_test if y_test is None else y_test

        scaler_y = StandardScaler()
        self.y_train_standardized_columnwise = scaler_y.fit_transform(y_train)
        self.y_val_standardized_columnwise = scaler_y.transform(y_val)
        self.y_test_standardized_columnwise = scaler_y.transform(y_test)
        self.standardize_y_ColumnWise = scaler_y

        self.LoadSave.load_or_dump_trained_object(
                                trained_object=self.standardize_y_ColumnWise,
                                output_indicator=output_indicator,
                                load_or_dump='dump')

        if print_model:
            print(scaler_y)


    def standardize_X_column_wise(self,
                                  output_indicator = 'Trained_StandardScaler_X_ColWise',
                                  X_train=None,
                                  X_val=None,
                                  X_test=None,
                                  print_model=False):
        """
        Standardize feature variables (X) column-wise by removing the mean and scaling to unit variance.
        Transform the data such that each feature will have a mean value of 0 and a standard deviation of 1.

        Parameters
        -----------
        X_train : array
            Training feature matrix.
        X_val : array
            Validation feature matrix.
        X_test : array
            Test feature matrix.
        print_model : bool
            Whether to print the trained scaler model.

        Returns
        -------
        self.X_train_standardized_columnwise : array
            Standardized training feature matrix.
        self.X_val_standardized_columnwise : array
            Standardized validation feature matrix.
        self.X_test_standardized_columnwise : array
            Standardized test feature matrix.
        """

        # Set default values if None is provided
        X_train = self.X_train if X_train is None else X_train
        X_val = self.X_val if X_val is None else X_val
        X_test = self.X_test if X_test is None else X_test

        scaler_X = StandardScaler()

        self.X_train_standardized_columnwise = scaler_X.fit_transform(X_train)
        self.X_val_standardized_columnwise = scaler_X.transform(X_val)
        self.X_test_standardized_columnwise = scaler_X.transform(X_test)

        self.standardize_X_ColumnWise = scaler_X

        self.LoadSave.load_or_dump_trained_object(
            trained_object=self.standardize_X_ColumnWise,
            output_indicator=output_indicator,
            load_or_dump='dump')

        if print_model:
            print(scaler_X)

    def standardize_X_row_wise(self,
                               output_indicator = 'Trained_StandardScaler_X_RowWise',
                               X_train=None,
                               X_val=None,
                               X_test=None,
                               print_model=False):
        """
        Standardize feature variables (X) column-wise by removing the mean and scaling to unit variance.
        Transform the data such that each feature will have a mean value of 0 and a standard deviation of 1.

        Parameters
        ----------
        X_train : array
            Training feature matrix.
        X_val : array
            Validation feature matrix.
        X_test : array
            Test feature matrix.
        print_model : bool
            Whether to print the trained scaler model.

        Returns
        -------
        self.X_train_standardized_columnwise : array
            Standardized training feature matrix.
        self.X_val_standardized_columnwise : array
            Standardized validation feature matrix.
        self.X_test_standardized_columnwise : array
            Standardized test feature matrix.
        """

        # Set default values if None is provided
        X_train = self.X_train if X_train is None else X_train
        X_val = self.X_val if X_val is None else X_val
        X_test = self.X_test if X_test is None else X_test

        scaler_X = StandardScaler()
        self.X_train_standardized_rowwise = scaler_X.fit_transform(X_train.T).T
        self.X_val_standardized_rowwise = scaler_X.fit_transform(X_val.T).T
        self.X_test_standardized_rowwise = scaler_X.fit_transform(X_test.T).T
        self.standardize_X_RowWise = scaler_X

        self.LoadSave.load_or_dump_trained_object(
            trained_object=self.standardize_X_RowWise,
            output_indicator=output_indicator,
            load_or_dump='dump')

        if print_model:
            print(scaler_X)

    def plot_boxplot_scaled_features(self, scaled_feature, title=None, xticks_list=None, fig_size=(14, 3)):
        """
        Make a boxplot with the scaled features.

        Description
        -----------
            - Median: middle quartile marks.
            - Inter-quartile range (The middle “box”): 50% of scores fall within the inter-quartile range.
            - Upper quartile: 75% of the scores fall below the upper quartile.
            - Lower quartile: 25% of scores fall below the lower quartile.
        """

        plt.figure(figsize=fig_size)
        plt.boxplot(scaled_feature, sym='')

        if len(scaled_feature) > 10:
            plt.xticks(rotation=45)

        plt.xlabel('Features', fontsize=12)
        plt.ylabel('Scaled Value', fontsize=12)
        if title:
            plt.title(title, fontsize=14)

        # Add custom x-ticks
        # custom_xticks = ['Label 1', 'Label 2', 'Label 3', 'Label 4']
        if xticks_list:
            xtick_positions = range(len(xticks_list))
            plt.xticks(xtick_positions, xticks_list)

        plt.tight_layout()
        plt.show()

    def plot_histogram_scaled_features(self, scaled_feature):
        """
        Plotting the histogram of scaled features
        """

        plt.figure(figsize=(15, 6))
        plt.hist(scaled_feature, bins=31)
        plt.xlabel('Scaled Value')
        plt.ylabel('Frequency')
        plt.title('Histogram of Scaled Features')
        plt.show()




# ======= Import functions/Classes from other modules ====================

from .IO_utils import *

from .StatVisAnalyzer import boxplot_hist, plot_spectra_errorbar, \
    plot_pred_vs_obs_errorbar
from .StatVisAnalyzer import interpolate_df, print_results_fun
from .StatVisAnalyzer import replace_zeros_with_mean, calculate_confidence_intervals_std_df, \
    plot_pred_vs_obs_errorbar_stat

# ======= Import Python libraries ========================================

# Standard Data Manipulation / Statistical Libraries ******
# import os
import pandas as pd
import numpy as np
# import pickle as pk
# from scipy import stats
# from uncertainties import ufloat

import astropy.units as u
from scipy.interpolate import pchip
# from tensorflow.keras.models import Sequential, model_from_json
import spectres

# Data Visulaization Libraries ****************************
import matplotlib.pyplot as plt
# import seaborn as sns

from bokeh.io import output_notebook

output_notebook()
from bokeh.plotting import show, figure

TOOLTIPS = [
    ("index", "$index"),
    ("(x,y)", "($x, $y)"),
]

# Data science / Machine learning Libraries ***************

# from astropy.constants import c
from astropy.nddata import StdDevUncertainty, NDDataArray
from bokeh.palettes import viridis


__reference_data__ = os.getenv("TelescopeML_reference_data")
# print(__reference_data__)


# ===============================================================================
# ========================                           ============================
# ========================    observational_spectra  ============================
# ========================                           ============================
# ===============================================================================


class ObserveParameterPredictor:
    """
    Load, process, and visualize observational spectra.

    Parameters
    ----------
    F_lambda_obs : array
        Fluxes for each feature (wavelength) from observational data.
    F_lambda_obs_err : array
        Observed spectra error bars.
    wl_obs : array
        Name of features (wavelength) from observational data, e.g., 0.9, 1.0, 1.1 micron.
    wl_synthetic : array
        Name of features (wavelengths) from synthetic data.

    Example
    -------
        >>> HD3651B_BD_literature_info = {
        >>>   'bd_name':'HD3651B',
        >>>   'bd_Teff':818,
        >>>   'bd_logg':3.94,
        >>>   'bd_met': -0.22,
        >>>   'bd_distance_pc' : 11.134,
        >>>   'bd_radius_Rjup' : 0.81,
        >>>   'bd_radius_Rjup_tuned': .81}
    """

    def __init__(self,
                 object_name,
                 training_dataset_df,
                 wl_synthetic,
                 bd_literature_dic,
                 trained_ML_model,
                 trained_X_ColWise_MinMax = None,
                 trained_y_ColWise = None,
                 trained_X_RowWise = None,
                 ):

        self.object_name = object_name
        self.training_dataset_df = training_dataset_df
        self.wl_synthetic = wl_synthetic
        self.bd_literature_dic = bd_literature_dic
        self.trained_ML_model = trained_ML_model
        self.trained_X_ColWise_MinMax = trained_X_ColWise_MinMax, #Trained_StandardScaler_X_ColWise_MinMax
        self.trained_y_ColWise = trained_y_ColWise, #Trained_StandardScaler_y_ColWise
        self.trained_X_RowWise = trained_X_RowWise, #Trained_StandardScaler_X_RowWise

        # print(self.trained_X_ColWise_MinMax)

    def load_observational_spectra(self,
                                   obs_data_df = None,
                                   __plot_observational_spectra_errorbar__=False,
                                   __replace_zeros_negatives_with_mean__=True,
                                   __print_results__=False,
                                   ):
        """
        Load the observational spectra, process the dataset, and optionally plot the observational spectra with error bars.

        Parameters
        -----------
        __plot_observational_spectra_errorbar__ : bool
            True or False.
        __replace_zeros_negatives_with_mean__ : bool
            True or False.
        __print_results__ : bool
            True or False.
        """

        # Load the observational spectra
        if obs_data_df is None:
            obs_data_df = pd.read_csv(__reference_data__+f'/observational_datasets/{self.object_name}_fluxcal.dat',
                                      delim_whitespace=True, comment='#', names=('wl', 'F_lambda', 'F_lambda_error'),
                                      usecols=(0, 1, 2))

        # # Process the dataset - remove Nan and negative values
        # obs_data_df['F_lambda'] = obs_data_df['F_lambda'].mask(obs_data_df['F_lambda'].lt(0), 0)
        # obs_data_df['F_lambda'].replace(0, np.nan, inplace=True)
        # obs_data_df['F_lambda'].interpolate(inplace=True)

        if __replace_zeros_negatives_with_mean__:
            obs_data_df['F_lambda_obs'] = obs_data_df['F_lambda'].mask(obs_data_df['F_lambda'].lt(0), 0)
            obs_data_df['F_lambda_obs'].replace(0, np.nan, inplace=True)
            obs_data_df['F_lambda_obs'].interpolate(method='linear', limit_direction='both', inplace=True)

            obs_data_df['F_lambda_obs_err'] = obs_data_df['F_lambda_error'].mask(obs_data_df['F_lambda_error'].lt(0), 0)
            obs_data_df['F_lambda_obs_err'].replace(0, np.nan, inplace=True)
            obs_data_df['F_lambda_obs_err'].interpolate(method='linear', limit_direction='both', inplace=True)

        self.obs_data_df = obs_data_df

        if __print_results__:
            print('------- Observational DataFrame Example ---------')
            print(self.obs_data_df.head(5))

        if __plot_observational_spectra_errorbar__:
            plot_spectra_errorbar(
                object_name=self.object_name,
                x_obs=self.obs_data_df['wl'],
                y_obs=self.obs_data_df['F_lambda_obs'],
                y_obs_err=self.obs_data_df['F_lambda_obs_err'],
                y_label='Flux (F𝛌) [erg/s/A/cm2]',
            )


    def ProcessObservationalDataset(self,
                                    F_lambda_obs,
                                    F_lambda_obs_err,
                                    wl_obs,
                                    # wl_synthetic,
                                    bd_literature_dic=None):
        """
        Process the observational dataset and set various attributes of the object.

        Parameters
        -----------
        F_lambda_obs : array
            Observed feature values.
        F_lambda_obs_err : array
            Errors corresponding to the observed feature values.
        wl_obs : array
            Names of the observed features.
        wl_synthetic : array
            Names of the synthetic features.
        bd_literature_dic : dict, optional
            Dictionary containing literature information. Defaults to None.

        Return
        -------
        Fnu_obs , Fnu_obs_err, Fnu_obs_absolute, Fnu_obs_absolute_err

        Example
        -------

        >>> HD3651B_BD_literature_info = {'bd_name':'HD3651B',
        >>>   'bd_Teff':818,
        >>>   'bd_logg':3.94,
        >>>   'bd_met': -0.22,
        >>>   'bd_distance_pc' : 11.134,
        >>>   'bd_radius_Rjup' : 0.81,
        >>>   'bd_radius_Rjup_tuned': .81}

        """

        self.F_lambda_obs = F_lambda_obs
        self.F_lambda_obs_err = F_lambda_obs_err
        self.wl_obs = wl_obs
        # self.wl_synthetic = wl_synthetic
        self.bd_literature_dic = bd_literature_dic

        Fnu_obs , Fnu_obs_err = self.Flam_to_Fnu(
                 Flam_values = self.F_lambda_obs,
                 Flam_errors = self.F_lambda_obs_err,
                 wavelengths = wl_obs,)  # Convert F_lambda_obs to Fnu_obs

        Fnu_obs_absolute, Fnu_obs_absolute_err = self.Fnu_to_Fnu_abs(
                 Fnu_values = Fnu_obs,
                 Fnu_errors = Fnu_obs_err,
                 bd_literature_dic = bd_literature_dic)


        return Fnu_obs , Fnu_obs_err, Fnu_obs_absolute, Fnu_obs_absolute_err


    def Fnu_to_Fnu_abs(self,
                       Fnu_values,
                       Fnu_errors,
                       bd_literature_dic = None,
                       __plot__ = False
                       ):

        # Calculate the absolute Fnu_obs values
        bd_literature_dic = self.bd_literature_dic if bd_literature_dic is None else bd_literature_dic

        Fnu_values_abs = Fnu_values * (bd_literature_dic['bd_distance_pc'] * ((u.pc).to(u.jupiterRad)) / (
        bd_literature_dic['bd_radius_Rjup'])) ** 2

        Fnu_errors_abs = Fnu_errors * (
                    bd_literature_dic['bd_distance_pc'] * ((u.pc).to(u.jupiterRad)) / (
            bd_literature_dic['bd_radius_Rjup'])) ** 2

        if __plot__:
            plot_spectra_errorbar(
                object_name=self.object_name,
                x_obs=self.wl_obs,
                y_obs=Fnu_values_abs,
                y_obs_err=Fnu_errors_abs,
                y_label='Abs. Flux (F𝜈) [erg/s/cm2/Hz]',
            )

        return Fnu_values_abs, Fnu_errors_abs



    def Flam_to_Fnu(self,
                 Flam_values,
                 Flam_errors,
                 wavelengths):
        """
        Convert F_lambda to F_nu along with error propagation.
        Parameters
        ----------
        Flam_values
        Flam_errors
        wavelengths

        Returns
        -------
        fnu_values : array
            Array of flux density values in F_nu.
        fnu_errors : array
            Array of error bars for the flux density values in F_nu.
        """

        # flam_values = self.F_lambda_obs
        # flam_errors = self.F_lambda_obs_err
        # wavelengths = self.wl_obs

        flam_dataset = NDDataArray(Flam_values, uncertainty=StdDevUncertainty(Flam_errors),
                                   unit=(u.erg / u.s / u.Angstrom / u.cm ** 2))

        Fnu_all = flam_dataset.convert_unit_to(unit=(u.erg / u.s / u.cm ** 2 / u.Hz),
                                               equivalencies=u.spectral_density(wavelengths * u.micron))
        Fnu_values = Fnu_all.data
        Fnu_errors = Fnu_all.uncertainty.array

        return Fnu_values, Fnu_errors

    def flux_interpolated(self,
                          Fnu_obs_absolute,
                          interpolated_wl = None,
                          __print_results__=False,
                          __plot_spectra_errorbar__=False,
                          __use_spectres__=True):
        """
        Perform flux interpolation using either SpectRes or pchip interpolation.

        Parameters
        ----------
        __print_results__ : bool
            True or False.
        __plot_spectra_errorbar__ : bool
            True or False.
        __use_spectres__ : bool, optional
            Whether to use SpectRes for interpolation. Defaults to True.

        Return
        -------
        Fnu_obs_absolute_intd, Fnu_obs_absolute_intd_df
        """
        Fnu_obs_absolute = self.Fnu_obs_absolute if Fnu_obs_absolute is None else Fnu_obs_absolute
        interpolated_wl = np.sort(self.wl_synthetic.wl.values) if interpolated_wl is None else interpolated_wl

        if __use_spectres__:
            Fnu_obs_absolute_intd = spectres.spectres(interpolated_wl,
                                                           np.float64(self.wl_obs),
                                                           np.float64(Fnu_obs_absolute))
        else:
            flux_intd = pchip(self.wl_obs, self.Fnu_obs_absolute)
            Fnu_obs_absolute_intd = flux_intd(interpolated_wl)

        Fnu_obs_absolute_intd_df = pd.DataFrame(list(Fnu_obs_absolute_intd),
                                                     index=[str(x) for x in np.round(interpolated_wl, 3)]).T


        if __print_results__:
            print('---    Object Flux     ----')
            print(self.Fnu_obs_absolute_intd_df)
            print('---    Object Flux Interpolate     ----')
            print(pd.DataFrame(self.Fnu_obs_absolute_intd))

        if __plot_spectra_errorbar__:
            plot_spectra_errorbar(
                object_name=self.object_name,
                x_obs=self.obs_data_df['wl'],
                y_obs=self.obs_data_df['F_lambda_obs'],
                y_obs_err=self.obs_data_df['F_lambda_obs_err'],
                y_label='Flux (F𝛌)',
            )

            plot_spectra_errorbar(
                object_name=self.object_name,
                x_obs=self.obs_data_df['wl'],
                y_obs=self.Fnu_obs,
                y_obs_err=self.Fnu_obs_err,
                y_label='Flux (F𝜈)',
            )

            plot_spectra_errorbar(
                object_name=self.object_name,
                x_obs=self.obs_data_df['wl'],
                y_obs=self.Fnu_obs_absolute,
                y_obs_err=self.Fnu_obs_absolute_err,
                y_label='Flux_abs (F_abs)',
            )


        return Fnu_obs_absolute_intd, Fnu_obs_absolute_intd_df




    def Process_Observational_Dataset(self,
                                      __print_results__=False,
                                      F_lambda_obs = None,
                                      F_lambda_obs_err = None,
                                      # __plot_predicted_vs_observed__=False
                                      ):
        """
        Process the observational dataset, extract ML features, perform predictions, and optionally print the results and plot the predicted versus observed spectra.

        Parameters
        ----------
        __print_results__ : bool
            True or False.
        """


        F_lambda_obs = self.obs_data_df['F_lambda_obs'] if F_lambda_obs == None else F_lambda_obs
        F_lambda_obs_err = self.obs_data_df['F_lambda_obs_err'] if F_lambda_obs_err == None else F_lambda_obs_err

        Fnu_obs , Fnu_obs_err, Fnu_obs_absolute, Fnu_obs_absolute_err = self.ProcessObservationalDataset(
                    F_lambda_obs=F_lambda_obs.values,
                    F_lambda_obs_err=F_lambda_obs_err.values,
                    wl_obs=self.obs_data_df['wl'].values,
                    # wl_synthetic=self.wl_synthetic['wl'].values,
                    bd_literature_dic=self.bd_literature_dic,
                   )

        # self.Fnu_obs, Fnu_obs_err, self.Fnu_obs_absolute, self.Fnu_obs_absolute_err = Fnu_obs , Fnu_obs_err, Fnu_obs_absolute, Fnu_obs_absolute_err
        self.obs_data_df['Fnu_obs'] = Fnu_obs
        self.obs_data_df['Fnu_obs_err'] = Fnu_obs_err
        self.obs_data_df['Fnu_obs_absolute'] = Fnu_obs_absolute
        self.obs_data_df['Fnu_obs_absolute_err'] = Fnu_obs_absolute_err

        # Extract the original ML features from the observational spectrum
        self.Fnu_obs_absolute_intd, self.Fnu_obs_absolute_intd_df = \
                    self.flux_interpolated(Fnu_obs_absolute = Fnu_obs_absolute,
                                           interpolated_wl = None,
                                           __print_results__=False,
                                           __plot_spectra_errorbar__=False,
                                           __use_spectres__=True)

        if __print_results__:
            print('------------  Interpolated Observational Spectra: Absolute F𝜈 ------------')
            print(self.Fnu_obs_absolute_intd_df)

        # Extract the engineered ML features from the observational spectrum
        df_Fnu_obs_absolute_intd_min = self.Fnu_obs_absolute_intd_df.min(axis=1)
        df_Fnu_obs_absolute_intd_max = self.Fnu_obs_absolute_intd_df.max(axis=1)

        self.df_MinMax_obs = pd.DataFrame((df_Fnu_obs_absolute_intd_min, df_Fnu_obs_absolute_intd_max)).T

        if __print_results__:
            print('------------ df_MinMax Single Observational Spectrum ------------')
            print(self.df_MinMax_obs)

        # this is commited b/c it is for standardize_X_ColumnWise for the MinMax
        XminXmax_Stand = self.trained_X_ColWise_MinMax[0].transform(self.df_MinMax_obs.values)

        # XminXmax_Stand = self.trained_data_processor.normalize_X_ColumnWise.transform(self.df_MinMax_obs.values)

        bd_mean = self.Fnu_obs_absolute_intd_df.mean(axis=1)[0]
        bd_std = self.Fnu_obs_absolute_intd_df.std(axis=1)[0]

        X_Scaled = (self.Fnu_obs_absolute_intd_df.values[0] - bd_mean) / bd_std


        y_pred_train = np.array(
            self.trained_ML_model.predict([X_Scaled[::-1].reshape(1, 104), XminXmax_Stand],
                                                                  verbose=0))[:, :, 0].T
        y_pred_train_ = self.trained_y_ColWise[0].inverse_transform(y_pred_train)
        y_pred_train_[:, 3] = 10 ** y_pred_train_[:, 3]
        y_pred = y_pred_train_
        self.y_pred = y_pred
        self.targets_single_spectrum_dic = dict(zip(['gravity', 'c_o_ratio', 'metallicity', 'temperature'],
                                                    [np.round(self.y_pred[0][i], 2) for i in range(4)]))
        # self.bd_object = bd_object

        if __print_results__:
            print_results_fun(targets=dict(zip(['bd_radius_Rjup', 'BD_mean', 'BD_std'],
                                               [self.bd_literature_dic['bd_radius_Rjup'], bd_mean, bd_std])),
                              print_title='Radius, BD_mean, BD_std:')

            # print('Radius, BD_mean, BD_std:',self.bd_literature_dic['bd_radius_Rjup'], bd_mean, bd_std)

            print_results_fun(targets=self.targets_single_spectrum_dic,
                              print_title='Predicted Targets from the Signle Observational Spectrum:')


    def predict_from_random_spectra(
            self,
            random_spectra_num=10,
            __print_results__=False,
            __plot_randomly_generated_spectra__=False,
            __plot_histogram__=False,
            __plot_boxplot_hist__=False,
            # __plot_predicted_vs_observed__=False,
            __plot_pred_vs_obs_errorbar__=False,
            __plot_pred_vs_obs_errorbar_stat__=False,
            __calculate_confidence_intervals_std_df__=False,
            ):
        """

        Generate random spectra based on the observational data, predict the target features, and plot the spectra

        Parameters
        ----------
        random_spectra_num : int, optional
            Number of random spectra to generate. Defaults to 10.
        __print_results__ : bool
            True or False.
        __plot_randomly_generated_spectra__ : bool
            True or False.
        __plot_histogram__ : bool
            True or False.
        __plot_boxplot_hist__ : bool
            True or False.
        __plot_pred_vs_obs_errorbar__ : bool
            True or False.
        __plot_pred_vs_obs_errorbar_stat__ : bool
            True or False.
        __calculate_confidence_intervals_std_df__ : bool
            True or False.
        """

        color = viridis(250).__iter__()

        spectra_list_obs = []
        spectra_list_pre = []
        param_list = []

        # Generate random spectra
        for i in range(random_spectra_num):
            spectra = pd.DataFrame(
                np.random.normal(self.obs_data_df['F_lambda_obs'], self.obs_data_df['F_lambda_obs_err']),
                columns=['F_lambda_obs']
            )

            # Process the dataset: fix negatives and nans
            spectra['F_lambda_obs'] = spectra['F_lambda_obs'].mask(spectra['F_lambda_obs'].lt(0), 0)
            spectra['F_lambda_obs'].replace(0, np.nan, inplace=True)
            spectra['F_lambda_obs'].interpolate(inplace=True)

            # Process the randomly generated Observational spectra
            Fnu_obs , Fnu_obs_err, Fnu_obs_absolute, Fnu_obs_absolute_err = self.ProcessObservationalDataset(
                    F_lambda_obs=spectra['F_lambda_obs'].values,
                    F_lambda_obs_err=self.obs_data_df['F_lambda_obs_err'].values,
                    wl_obs=self.obs_data_df['wl'].values,
                    # wl_synthetic=self.wl_synthetic['wl'].values,
                    bd_literature_dic=self.bd_literature_dic,
                    )

            # self.obs_data_df['Fnu_obs'] = self.Fnu_obs
            # self.obs_data_df['Fnu_obs_err'] = self.Fnu_obs_err
            # self.obs_data_df['Fnu_obs_absolute'] = self.Fnu_obs_absolute
            # self.obs_data_df['Fnu_obs_absolute_err'] = self.Fnu_obs_absolute_err

            # Extract the original ML features from the observational spectrum
            Fnu_obs_absolute_intd, Fnu_obs_absolute_intd_df = \
                        self.flux_interpolated(Fnu_obs_absolute = Fnu_obs_absolute,
                                               interpolated_wl = None,
                                               __print_results__=False,
                                               __plot_spectra_errorbar__=False,
                                               __use_spectres__=True)

            # Extract the engineered ML features from the observational spectrum
            Fnu_obs_absolute_intd_df_min = Fnu_obs_absolute_intd_df.min(axis=1)
            Fnu_obs_absolute_intd_df_max = Fnu_obs_absolute_intd_df.max(axis=1)

            df_MinMax_obs = pd.DataFrame(
                (Fnu_obs_absolute_intd_df_min, Fnu_obs_absolute_intd_df_max)
            ).T

            XminXmax_Stand = self.trained_X_ColWise_MinMax[0].transform(df_MinMax_obs.values)
            # XminXmax_Stand = self.trained_data_processor.normalize_X_ColumnWise.transform(df_MinMax_obs.values)


            bd_mean = Fnu_obs_absolute_intd_df.mean(axis=1)[0]
            bd_std = Fnu_obs_absolute_intd_df.std(axis=1)[0]


            # X_Scaled = (Fnu_obs_absolute_intd_df.div((self.bd_literature_dic['bd_radius_Rjup'])**2).values[0] - bd_mean) / bd_std
            X_Scaled = (Fnu_obs_absolute_intd_df.values[0] - bd_mean) / bd_std

            y_pred_train = np.array(
                self.trained_ML_model.predict(
                    [X_Scaled[::-1].reshape(1, 104), XminXmax_Stand.reshape(1, 2)], verbose=0) #findme!
            )[:, :, 0].T

            y_pred_train_ = self.trained_y_ColWise[0].inverse_transform(y_pred_train)
            y_pred_train_[:, 3] = 10 ** y_pred_train_[:, 3]
            y_pred_random = y_pred_train_

            targets_dic_random = dict(zip(['gravity', 'c_o_ratio', 'metallicity', 'temperature'],
                                          [y_pred_random[0][i] for i in range(4)])
                                      )


            spectra_list_obs.append(Fnu_obs_absolute_intd_df.values)
            param_list.append(y_pred_random[0])

            filtered_df4 = interpolate_df(dataset=self.training_dataset_df,
                                          predicted_targets_dic=targets_dic_random,
                                          # self.targets_single_spectrum_dic,#
                                          print_results_=False)
            # self.filtered_df4 = filtered_df4
            # print(filtered_df4.iloc[0,0:-5].values)

            spectra_list_pre.append(filtered_df4.iloc[:, 0:-5].div((self.bd_literature_dic['bd_radius_Rjup'])**2).values.flatten())
            # spectra_list_pre.append(filtered_df4.iloc[:, 0:-5].values.flatten())


        self.spectra_list_obs = spectra_list_obs
        self.spectra_list_pre = spectra_list_pre
        self.param_list = param_list

        self.df_random_pred = pd.DataFrame(self.param_list, columns=['logg', 'c_o', 'met', 'T'])

        self.dic_random_pred_mean = dict(
            zip(['gravity', 'c_o_ratio', 'metallicity', 'temperature'],
                list(self.df_random_pred.agg(np.mean)))
        )

        # self.df_spectra_list_obs = pd.DataFrame(data=np.array(self.spectra_list_obs).reshape(-1, 104), columns=self.wl_synthetic.wl_synthetic)
        self.df_spectra_list_pre = pd.DataFrame(data=self.spectra_list_pre, columns=self.wl_synthetic.wl[::-1])
        # print(self.spectra_list_pre)
        # print(self.df_spectra_list_obs)

        if __print_results__:
            print_results_fun(targets=self.dic_random_pred_mean,
                              print_title='Predicted Targets from Randomly Generated Spectra:')

        if __print_results__:
            print(self.df_random_pred.describe())

        if __plot_randomly_generated_spectra__:
            p = figure(
                title=self.object_name + ": Randomly generated spectra within 1σ",
                x_axis_label='Features (Wavelength [𝜇m])',
                y_axis_label='Flux (F𝜈) [erg/s/cm2/Hz]',
                width=800,
                height=300,
                y_axis_type="log",
                background_fill_color="#fafafa",
            )

            if random_spectra_num > 250:
                random_spectra_num_index = 200
            else:
                random_spectra_num_index = random_spectra_num

            for i in range(0, random_spectra_num_index):
                p.line(
                    self.wl_synthetic.wl.values[::-1],
                    self.spectra_list_obs[i][0],
                    line_width=1,
                    line_alpha=0.6,
                    line_color=next(color),
                )

            # Increase size of x and y ticks
            p.title.text_font_size = '12pt'
            p.xaxis.major_label_text_font_size = '12pt'
            p.xaxis.axis_label_text_font_size = '12pt'
            p.yaxis.major_label_text_font_size = '12pt'
            p.yaxis.axis_label_text_font_size = '12pt'

            show(p)

        if __plot_histogram__:
            plt.figure()
            self.df_random_pred.hist()
            plt.show()

        if __plot_boxplot_hist__:
            boxplot_hist(self.df_random_pred['logg'], x_label=r'$\log g$', xy_loc=[0.05, 0.98])
            boxplot_hist(self.df_random_pred['T'], x_label=r'$T_{eff}$', xy_loc=[0.05, 0.98])
            boxplot_hist(self.df_random_pred['c_o'], x_label=r'C/O', xy_loc=[0.05, 0.98])
            boxplot_hist(self.df_random_pred['met'], x_label=r'[M/H]', xy_loc=[0.05, 0.98])



        if __plot_pred_vs_obs_errorbar__:
            plot_pred_vs_obs_errorbar(
                object_name=self.object_name,
                x_obs=self.obs_data_df['wl'],
                y_obs=Fnu_obs_absolute,
                y_obs_error=Fnu_obs_absolute_err,
                training_dataset=self.training_dataset_df,
                x_pred=self.wl_synthetic,
                predicted_targets_dic=self.dic_random_pred_mean,
                __print_results__=False,
            )

        if __calculate_confidence_intervals_std_df__:
            self.confidence_intervals_std_df = calculate_confidence_intervals_std_df(
                dataset_df=self.df_spectra_list_pre,
                __print_results__=False,
                __plot_calculate_confidence_intervals_std_df__=False,
            )

        if __plot_pred_vs_obs_errorbar_stat__:
            plot_pred_vs_obs_errorbar_stat(
                stat_df=self.confidence_intervals_std_df,
                confidence_level=0.95,
                object_name=self.object_name,
                x_obs=self.obs_data_df['wl'],
                y_obs=self.obs_data_df['Fnu_obs_absolute'], #self.obs_data_df['Fnu_obs_absolute'],# self.Fnu_obs_absolute,
                y_obs_err=self.obs_data_df['Fnu_obs_absolute_err'], #self.obs_data_df['Fnu_obs_absolute_err'],#self.Fnu_obs_absolute_err,
                training_datasets=self.training_dataset_df,
                x_pred=self.wl_synthetic,
                predicted_targets_dic=self.dic_random_pred_mean,  # self.dic_random_pred_mean,
                radius = self.bd_literature_dic['bd_radius_Rjup'],
                __print_results__=False,
            )



# Import functions/Classes from other modules ====================

# from io_funs import LoadSave

# Import libraries ========================================

# ******* Standard Data Manipulation / Statistical Libraries *****
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
# import pickle as pk

from scipy import stats
from sklearn.metrics import r2_score, mean_squared_error
from scipy.interpolate import RegularGridInterpolator
from scipy.stats import chi2
import os

import pprint

# ******* Data Visulaization Libraries ****************************
import seaborn as sns
import matplotlib.pyplot as plt


from bokeh.plotting import output_notebook

from bokeh.plotting import figure, show
from bokeh.models import ColumnDataSource


def print_results_fun(targets, print_title=None):
    """
    Print the outputs in a pretty format using the pprint library.

    Parameters
    ----------
    targets : any
        The data to be printed.
    print_title : str
        An optional title to display before the printed data.
    """

    print('*' * 30 + '\n')

    if print_title is not None:
        print(print_title+ '\n')

    # Use pprint to print the data in a well-formatted and indented manner.
    pprint.pprint(targets, indent=4, width=30)

    print('*' * 30 + '\n')



def regression_report(trained_model,
                      trained_data_processor,
                      Xtrain, Xtest, ytrain, ytest,
                      target_i,
                      xy_top=None, xy_bottom=None, __print_results__= False):
    """
    Generate a regression report for the trained ML/CNN model.

    Parameters
    -----------
    trained_model : object
        Trained regression model.
    Xtrain : array
        Training set.
    Xtest : array
        Test set.
    ytrain : array
        Training target set.
    ytest : array
        Test target set.
    target_i : int
        Index of the target variable to analyze.
    xy_top : list, optional
        Coordinates for annotations in the top plot. Defaults to [0.55, 0.85].
    xy_bottom : list, optional
        Coordinates for annotations in the bottom plot. Defaults to [0.05, 0.8].
    __print_results__ : bool, optional
        True or False.
    """

    # Apply the trained ML model on the train set to predict the targets
    if xy_bottom is None:
        xy_bottom = [0.05, 0.8]
    if xy_top is None:
        xy_top = [0.55, 0.85]
    y_pred_train = np.array(trained_model.predict(Xtrain))[:, :, 0].T
    y_pred_train_list = trained_data_processor.standardize_y_ColumnWise.inverse_transform(y_pred_train)
    y_pred_train_list[:, 3] = 10 ** y_pred_train_list[:, 3]

    y_act_train_list = trained_data_processor.standardize_y_ColumnWise.inverse_transform(ytrain)
    y_act_train_list[:, 3] = 10 ** y_act_train_list[:, 3]

    # Apply the trained ML model on the test set to predict the targets
    y_pred_test = np.array(trained_model.predict(Xtest))[:, :, 0].T
    y_pred_test_list = trained_data_processor.standardize_y_ColumnWise.inverse_transform(y_pred_test)
    y_pred_test_list[:, 3] = 10 ** y_pred_test_list[:, 3]

    y_act_test_list = trained_data_processor.standardize_y_ColumnWise.inverse_transform(ytest)
    y_act_test_list[:, 3] = 10 ** y_act_test_list[:, 3]

    for i in range(0, target_i):
        y_pred_train = y_pred_train_list[:, i]
        y_act_train = y_act_train_list[:, i]
        y_pred_test = y_pred_test_list[:, i]
        y_act_test = y_act_test_list[:, i]

        # Calculate the residual (Predicted - Actual)
        residual_train_list = y_pred_train - y_act_train
        residual_test_list = y_pred_test - y_act_test

        # Calculate mean and standard deviation for residuals
        mean_test = np.round(np.mean(residual_test_list), 2)
        std_test = np.round(np.std(residual_test_list), 2)
        mean_train = np.round(np.mean(residual_train_list), 2)
        std_train = np.round(np.std(residual_train_list), 2)

        # Calculate skewness for residuals
        skew_test = stats.skew(residual_test_list)
        skew_train = stats.skew(residual_train_list)

        # Calculate R-squared scores
        r2_score_train = r2_score(y_pred_train, y_act_train)
        r2_score_test = r2_score(y_pred_test, y_act_test)

        # Calculate  RMSE scores
        rmse_score_train = np.sqrt(mean_squared_error(y_pred_train, y_act_train))
        rmse_score_test = np.sqrt(mean_squared_error(y_pred_test, y_act_test))


        # Create subplots for histograms and scatter plots
        f, axs = plt.subplots(2, 1, figsize=(5, 5), sharey=False, sharex=False,
                              gridspec_kw=dict(height_ratios=[1, 3]))

        if __print_results__:
            print('\n\n----------------------- Test ------------------------')
            print('R2: {:2.2f} \t  RMSE: {:2.2f} \t Mean+/-STD: {:2.2f}+/-{:2.2f}'.format(
                r2_score_test, rmse_score_train, mean_test, std_test))

            print('\n----------------------- Train ------------------------')
            print('R2: {:2.2f} \t  RMSE: {:2.2f} \t Mean+/-STD: {:2.2f}+/-{:2.2f}'.format(
                r2_score_train, rmse_score_test, mean_train, std_train))

        # Plot histograms of residuals
        axs[0].set_title(['Gravity', 'C_O_ratio', 'Metallicity', 'Temperature'][i], fontsize=14)
        sns.histplot(data=residual_train_list, ax=axs[0], label='train', alpha=0.7, bins=19,
                     log_scale=False, stat='percent', legend=True, linewidth=0)
        sns.histplot(data=residual_test_list, label='test', ax=axs[0], alpha=0.3, bins=19,
                     stat='percent', legend=True, linewidth=0)
        axs[0].set_xlim((-(abs(mean_train) + 3 * std_train), (abs(mean_train) + 3 * std_train)))
        axs[0].set_ylim((1e-1, 100))
        axs[0].set_yscale('log')
        axs[0].set_ylabel('Probability %', fontsize=12)

        # Plot scatter figures of predicted vs actual values
        sns.scatterplot(y=y_pred_train, x=y_act_train, label='train', ax=axs[1], alpha=0.7, legend=False)
        sns.scatterplot(y=y_pred_test, x=y_act_test, label='test', ax=axs[1], alpha=0.7, legend=False)
        axs[1].set_ylabel('Predicted value', fontsize=12)
        axs[1].set_xlabel('Actual value', fontsize=12)

        # Add annotations for skewness and R-squared scores
        axs[0].annotate(r'$\tilde{\mu}_{{\rm 3, train}}$= ' + f'{np.round(skew_train, 2)}',
                        fontsize=11, xy=(xy_top[0], xy_top[1] + 0.08), xycoords='axes fraction')
        axs[0].annotate(r'$\tilde{\mu}_{{\rm 3, test}}$ = ' + f'{np.round(skew_test, 2)}',
                        fontsize=11, xy=(xy_top[0], xy_top[1] - 0.08), xycoords='axes fraction')
        axs[1].annotate(r'R$^2_{\rm train}$=' + f'{"%0.2f" % r2_score_train} [{"%0.2f" % abs(mean_train)}$\pm${"%0.2f" % std_train}]',
                        fontsize=11, xy=(xy_bottom[0], xy_bottom[1] + 0.06), xycoords='axes fraction')
        axs[1].annotate(r'R$^2_{\rm test}$ =' + f'{np.round(r2_score_test, 2)} [{"%0.2f" % mean_test}$\pm${"%0.2f" % std_test}]',
                        fontsize=11, xy=(xy_bottom[0], xy_bottom[1] - 0.06), xycoords='axes fraction')

        axs[1].legend(loc='lower right', fontsize=11)

        f.tight_layout()
        target_name = ['Gravity', 'C_O_ratio', 'Metallicity', 'Temperature'][i]
        # plt.savefig(f'../outputs/figures/regression_report_{target_name}.pdf', format='pdf')
        plt.show()

def filter_dataset_range(dataset, filter_params):
    """
    filer the dataframe
    """
    filtered_df = dataset.copy()

    for param, bounds in filter_params.items():
        lower_bound, upper_bound = bounds
        filtered_df = filtered_df[(filtered_df[param] >= lower_bound) & (filtered_df[param] <= upper_bound)]

    return filtered_df

def find_nearest_top_bottom(value, lst):
    """
    Find the nearest value in the list of data
    """
    lst.sort()
    nearest_top = None
    nearest_bottom = None

    for num in lst:
        if num >= value:
            nearest_top = num
            break

    if nearest_top is None:
        nearest_top = lst[-1]

    for num in reversed(lst):
        if num <= value:
            nearest_bottom = num
            break

    if nearest_bottom is None:
        nearest_bottom = lst[0]

    return nearest_bottom, nearest_top


def interpolate_df(dataset,
                   predicted_targets_dic,
                   print_results_=False):
    """
    Interpolate the training set.

    Parameters
    ----------
    dataset : array
        The training dataset to be interpolated.
    predicted_targets_dic : dict
        Target features to be interpolated.
    """

    my_list_g = list(dataset['gravity'].sort_values().unique())
    my_list_met = list(dataset['metallicity'].sort_values().unique())
    my_list_c_o = list(dataset['c_o_ratio'].sort_values().unique())
    my_list_T = list(dataset['temperature'].sort_values().unique())

    g0, g1 = find_nearest_top_bottom(predicted_targets_dic['gravity'], my_list_g)
    co0, co1 = find_nearest_top_bottom(predicted_targets_dic['c_o_ratio'], my_list_c_o)
    met0, met1 = find_nearest_top_bottom(predicted_targets_dic['metallicity'], my_list_met)
    T0, T1 = find_nearest_top_bottom(predicted_targets_dic['temperature'], my_list_T)

    filter_params = {'gravity': (g0, g1),
                     'temperature': (T0, T1),
                     'c_o_ratio': (co0, co1),
                     'metallicity': (met0, met1)}

    df_to_interpolate = filter_dataset_range(dataset, filter_params).reset_index(drop=True)

    my_list_g = list(df_to_interpolate['gravity'].sort_values().unique())
    my_list_met = list(df_to_interpolate['metallicity'].sort_values().unique())
    my_list_c_o = list(df_to_interpolate['c_o_ratio'].sort_values().unique())
    my_list_T = list(df_to_interpolate['temperature'].sort_values().unique())

    if print_results_:
        print(df_to_interpolate)
        print(my_list_g, my_list_c_o, my_list_T, my_list_met)

    df_interpolated_ = pd.DataFrame(columns=df_to_interpolate.drop(
        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).columns)

    df_interpolated_all = pd.DataFrame(columns=df_to_interpolate.columns)

    for temp in my_list_T:
        for grav in my_list_g:
            for met in my_list_met:
                for c_o in range(0, len(my_list_c_o) - 1):
                    # print(temp, grav, met, c_o)

                    filter_params = {'gravity': (grav, grav),
                                     'temperature': (temp, temp),
                                     'c_o_ratio': (my_list_c_o[c_o], my_list_c_o[c_o + 1]),
                                     'metallicity': (met, met)}

                    df_to_interpolate_ = filter_dataset_range(dataset, filter_params).reset_index(
                        drop=True)  # .drop_duplicates(subset=['gravity', 'temperature', 'c_o_ratio', 'metallicity'])

                    data = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented'])

                    y = df_to_interpolate_['c_o_ratio'].to_numpy()

                    column_grid = data.columns.to_numpy().astype(float)
                    values = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).to_numpy()

                    interp_func = RegularGridInterpolator((y, column_grid), values)

                    # Define the coordinates for interpolation
                    xi = column_grid  # x-coordinates for interpolation

                    yi = predicted_targets_dic['c_o_ratio']
                    # np.append(np.arange(y[0], y[1], abs(y[1] - y[0])/5, dtype=np.float64),y[1])  # y-coordinates for interpolation
                    xi_mesh, yi_mesh = np.meshgrid(xi, yi, indexing='ij')  # Meshgrid for interpolation

                    # Perform interpolation
                    df_interpolated_ = pd.DataFrame(interp_func((yi_mesh, xi_mesh)).T, columns=data.columns,
                                                    dtype=np.float64)
                    df_interpolated_['c_o_ratio'] = yi_mesh[0]
                    df_interpolated_['temperature'] = temp
                    df_interpolated_['metallicity'] = met
                    df_interpolated_['gravity'] = grav
                    df_interpolated_['is_augmented'] = 'no'

                    df_interpolated_all = pd.concat([df_interpolated_, df_interpolated_all], ignore_index=True)

    # ***************************************************************************************

    my_list_g = list(df_interpolated_all['gravity'].sort_values().unique())
    my_list_met = list(df_interpolated_all['metallicity'].sort_values().unique())
    my_list_c_o = list(df_interpolated_all['c_o_ratio'].sort_values().unique())
    my_list_T = list(df_interpolated_all['temperature'].sort_values().unique())

    df_interpolated_all2 = df_interpolated_all
    df_interpolated_all2.drop_duplicates(inplace=True)

    df_interpolated_ = pd.DataFrame(columns=df_interpolated_all2.drop(
        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).columns)

    for c_o in my_list_c_o:
        for temp in my_list_T:
            for grav in my_list_g:
                for met in range(0, len(my_list_met) - 1):

                    filter_params = {'gravity': (grav, grav),
                                     'temperature': (temp, temp),
                                     'c_o_ratio': (c_o, c_o),
                                     'metallicity': (my_list_met[met], my_list_met[met + 1])}

                    df_to_interpolate_ = filter_dataset_range(df_interpolated_all2, filter_params).reset_index(
                        drop=True)  # .drop_duplicates(subset=['gravity', 'temperature', 'c_o_ratio', 'metallicity'])

                    data = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented'])

                    y = df_to_interpolate_['metallicity'].to_numpy()

                    column_grid = data.columns.to_numpy().astype(float)
                    values = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).to_numpy()

                    interp_func = RegularGridInterpolator((y, column_grid), values)

                    # Define the coordinates for interpolation
                    xi = column_grid  # x-coordinates for interpolation

                    yi = predicted_targets_dic['metallicity']
                    xi_mesh, yi_mesh = np.meshgrid(xi, yi, indexing='ij')  # Meshgrid for interpolation

                    # Perform interpolation
                    df_interpolated_ = pd.DataFrame(interp_func((yi_mesh, xi_mesh)).T, columns=data.columns,
                                                    dtype=np.float64)
                    df_interpolated_['metallicity'] = yi_mesh[0]
                    df_interpolated_['temperature'] = temp
                    df_interpolated_['c_o_ratio'] = c_o
                    df_interpolated_['gravity'] = grav
                    df_interpolated_['is_augmented'] = 'no'

                    df_interpolated_all = pd.concat([df_interpolated_, df_interpolated_all], ignore_index=True)


    # ************************************************************************************

    my_list_g = list(df_interpolated_all['gravity'].sort_values().unique())
    my_list_met = list(df_interpolated_all['metallicity'].sort_values().unique())
    my_list_c_o = list(df_interpolated_all['c_o_ratio'].sort_values().unique())
    my_list_T = list(df_interpolated_all['temperature'].sort_values().unique())

    if print_results_:
        print(my_list_g, my_list_c_o, my_list_T, my_list_met)

    df_interpolated_all2 = df_interpolated_all
    df_interpolated_all2.drop_duplicates(inplace=True)

    df_interpolated_ = pd.DataFrame(columns=df_interpolated_all2.drop(
        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).columns)

    for c_o in my_list_c_o:
        for met in my_list_met:
            for grav in my_list_g:
                for temp in range(0, len(my_list_T) - 1):
                    # print(temp, grav, met, c_o)

                    filter_params = {'gravity': (grav, grav),
                                     'temperature': (my_list_T[temp], my_list_T[temp + 1]),
                                     'c_o_ratio': (c_o, c_o),
                                     'metallicity': (met, met)}

                    df_to_interpolate_ = filter_dataset_range(df_interpolated_all2, filter_params).reset_index(
                        drop=True)  # .drop_duplicates(subset=['gravity', 'temperature', 'c_o_ratio', 'metallicity'])

                    # print(df_to_interpolate)
                    data = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented'])

                    y = df_to_interpolate_['temperature'].to_numpy()
                    # print(y)

                    column_grid = data.columns.to_numpy().astype(float)
                    values = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).to_numpy()

                    interp_func = RegularGridInterpolator((y, column_grid), values)

                    # Define the coordinates for interpolation
                    xi = column_grid  # x-coordinates for interpolation

                    yi = predicted_targets_dic['temperature']

                    xi_mesh, yi_mesh = np.meshgrid(xi, yi, indexing='ij')  # Meshgrid for interpolation

                    # Perform interpolation
                    df_interpolated_ = pd.DataFrame(interp_func((yi_mesh, xi_mesh)).T, columns=data.columns,
                                                    dtype=np.float64)
                    df_interpolated_['temperature'] = yi_mesh[0]
                    df_interpolated_['metallicity'] = met
                    df_interpolated_['c_o_ratio'] = c_o
                    df_interpolated_['gravity'] = grav
                    df_interpolated_['is_augmented'] = 'no'

                    df_interpolated_all = pd.concat([df_interpolated_, df_interpolated_all], ignore_index=True)

    # ******************************************************************************************

    my_list_g = list(df_interpolated_all['gravity'].sort_values().unique())
    my_list_met = list(df_interpolated_all['metallicity'].sort_values().unique())
    my_list_c_o = list(df_interpolated_all['c_o_ratio'].sort_values().unique())
    my_list_T = list(df_interpolated_all['temperature'].sort_values().unique())

    df_interpolated_all2 = df_interpolated_all
    df_interpolated_all2.drop_duplicates(inplace=True)

    df_interpolated_ = pd.DataFrame(columns=df_interpolated_all2.drop(
        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).columns)

    for c_o in my_list_c_o:
        for met in my_list_met:
            for temp in my_list_T:
                for grav in range(0, len(my_list_g) - 1):
                    # print(temp, grav, met, c_o)

                    filter_params = {'gravity': (my_list_g[grav], my_list_g[grav + 1]),
                                     'temperature': (temp, temp),
                                     'c_o_ratio': (c_o, c_o),
                                     'metallicity': (met, met)}

                    df_to_interpolate_ = filter_dataset_range(df_interpolated_all2, filter_params).reset_index(
                        drop=True)  # .drop_duplicates(subset=['gravity', 'temperature', 'c_o_ratio', 'metallicity'])

                    data = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented'])

                    y = df_to_interpolate_['gravity'].to_numpy()

                    column_grid = data.columns.to_numpy().astype(float)
                    values = df_to_interpolate_.drop(
                        columns=['gravity', 'temperature', 'c_o_ratio', 'metallicity', 'is_augmented']).to_numpy()

                    interp_func = RegularGridInterpolator((y, column_grid), values)

                    # Define the coordinates for interpolation
                    xi = column_grid  # x-coordinates for interpolation

                    yi = predicted_targets_dic['gravity']

                    xi_mesh, yi_mesh = np.meshgrid(xi, yi, indexing='ij')  # Meshgrid for interpolation

                    # Perform interpolation
                    df_interpolated_ = pd.DataFrame(interp_func((yi_mesh, xi_mesh)).T, columns=data.columns,
                                                    dtype=np.float64)
                    df_interpolated_['gravity'] = yi_mesh[0]
                    df_interpolated_['metallicity'] = met
                    df_interpolated_['c_o_ratio'] = c_o
                    df_interpolated_['temperature'] = temp
                    df_interpolated_['is_augmented'] = 'no'

                    df_interpolated_all = pd.concat([df_interpolated_, df_interpolated_all], ignore_index=True)


    df_interpolated_all.drop_duplicates(inplace=True)

    df_interpolated_final = df_interpolated_all[
        (df_interpolated_all['temperature'] == predicted_targets_dic['temperature']) &
        (df_interpolated_all['c_o_ratio'] == predicted_targets_dic['c_o_ratio']) &
        (df_interpolated_all['metallicity'] == predicted_targets_dic['metallicity']) &
        (df_interpolated_all['gravity'] == predicted_targets_dic['gravity']) &
        (df_interpolated_all['is_augmented'] == 'no')]

    return df_interpolated_final


def filter_dataframe(training_datasets, predicted_targets_dic):
    nearest_value_list = []
    filtered_df = training_datasets.copy()

    # Check if the values exist in the respective columns
    for col, value in predicted_targets_dic.items():
        if value not in filtered_df[col].values:
            nearest_value = filtered_df[col].values[np.argmin(np.abs(filtered_df[col].values - value))]
            nearest_value_list.append(nearest_value)
            filtered_df = filtered_df[filtered_df[col] == nearest_value]

    return nearest_value_list, filtered_df


def boxplot_hist(data,
                 x_label,
                 xy_loc):

    fig, (ax_box, ax_hist) = plt.subplots(2, sharex=True, gridspec_kw={"height_ratios": (.15, .85)})

    sns.histplot(data, ax=ax_hist, kde=True, stat='probability')
    sns.boxplot(x = data, ax=ax_box, showmeans=True, meanline = True,
                meanprops={"marker": "|",
                           "markeredgecolor": "white",
                           "markersize": "30",
                            }
                       )

    fig.set_figheight(3)
    fig.set_figwidth(3)

    ax_box.set(xlabel='')
    sns.despine(ax=ax_hist)
    sns.despine(ax=ax_box, left=True)
    ax_box.set_yticks([])

    mean = np.round(np.mean(data),2)
    std = np.round(np.std(data),2)
    plt.annotate(f'{x_label}='+str(np.round(mean,2))+'$\pm$'+str(np.round(std,2)), fontsize=11,
                 xy=(xy_loc[0], xy_loc[1]), xycoords='axes fraction')

    plt.xlabel(x_label, fontsize = 12)
    if x_label == 'C/O':
        x_label = 'c_o_ratio'
    if x_label == '[M/H]':
        x_label = 'metallicity'
    # plt.savefig(f'../outputs/figures/boxplot_hist_{x_label}.pdf', format='pdf')

    plt.show()


def plot_spectra_errorbar(object_name,
                          x_obs,
                          y_obs,
                          y_obs_err,
                          y_label = "Flux (F𝜈) [erg/s/cm2/Hz]",
                          title_label = None,
                          data_type='x_y_yerr'):
    # Create the figure
    p = figure(title=f"{object_name}: Calibrated Observational Spectra" if title_label is None else title_label,
               x_axis_label="Features (Wavelength [𝜇m])",
               y_axis_label=y_label,
               width=800, height=300,
               y_axis_type="log",
               tools="pan,wheel_zoom,box_zoom,reset")



    # Add the scatter plot
    p.scatter(x_obs, y_obs,  size=4, fill_color='green', line_color=None, line_alpha=0.2,
              legend_label=f"{object_name}: Observational data")

    if data_type == 'x_y_yerr':
        # Define maximum error threshold as a percentage of y-value
        max_error_threshold = 0.8

        # Calculate adjusted error bar coordinates
        upper = np.minimum(y_obs + y_obs_err, y_obs + y_obs * max_error_threshold)
        lower = np.maximum(y_obs - y_obs_err, y_obs - y_obs * max_error_threshold)# Sample data
        p.segment(x0=x_obs, y0=lower, x1=x_obs, y1=upper,
                  color='gray', line_alpha=0.7)

    # Increase size of x and y ticks
    p.title.text_font_size = '12pt'
    p.xaxis.major_label_text_font_size = '12pt'
    p.xaxis.axis_label_text_font_size = '12pt'
    p.yaxis.major_label_text_font_size = '12pt'
    p.yaxis.axis_label_text_font_size = '12pt'

    # Show the plot
    output_notebook()
    show(p)


def replace_zeros_with_mean(dataframe_col):
    """
    Replace zero values in a DataFrame column with the mean of their non-zero neighbors.

    Parameters
    ----------
    dataframe_col : pandas.Series
        A pandas Series representing the column of a DataFrame.

    Returns
    -------
    pandas.Series
        The updated pandas Series with zero values replaced by the mean of non-zero neighbors.
    """

    df_col = dataframe_col.copy()
    zero_indices = np.where(df_col.values <= 0)
    non_zero_indices = np.where(df_col.values > 0)

    if zero_indices[0].size > 0:
        for row in zero_indices[0]:
            neighbors = df_col.iloc[max(0, row-1):row+2]
            neighbors = neighbors[neighbors != 0]
            while len(neighbors) == 0:
                row -= 1
                neighbors = df_col.iloc[max(0, row-1):row+2]
                neighbors = neighbors[neighbors != 0]
            df_col.iloc[row] = np.mean(neighbors)

        return df_col

    else:
        return dataframe_col



def plot_pred_vs_obs_errorbar(object_name,
                                       x_obs,
                                       y_obs,
                                       y_obs_error,
                                       training_dataset,
                                       x_pred,
                                       predicted_targets_dic,
                                       __print_results__= False):
    """
    Plot predicted spectra along with observed spectra and error bars.

    Parameters
    ----------
    object_name : str
        Name of the object being plotted.
    x_obs : list
        List of x-axis values (wavelengths) for the observed spectra.
    y_obs : list
        List of y-axis values (observed feature values) for the observed spectra.
    y_obs_error : list
        List of error values corresponding to the observed feature values.
    training_datasets : list, optional
        Training (or synthetic) datasets used for training the models. Default is None.
    predicted_targets_dic : dict, optional
        Dictionary of predicted targets. Default is None.
    __print_results__ : bool
        True or False.
    """

    # Define maximum error threshold as a percentage of y-value
    max_error_threshold = 0.8


    # Calculate adjusted error bar coordinates
    upper = np.minimum(y_obs + y_obs_error, y_obs + y_obs * max_error_threshold)
    lower = np.maximum(y_obs - y_obs_error, y_obs - y_obs * max_error_threshold)# Sample data


    # Create a ColumnDataSource to store the data
    source = ColumnDataSource(data=dict(x=x_obs, y=y_obs, upper=upper, lower=lower))

    # Create the Observational figure
    p = figure(title=f"{object_name}: Calibrated Observational VS. Predicted Spectra",
               x_axis_label="Features (Wavelength [𝜇m])",
               y_axis_label="Flux (F𝜈) [erg/s/cm2/Hz]",
               width=800, height=300,
               y_axis_type="log",
               tools="pan,wheel_zoom,box_zoom,reset")

    # Add the scatter plot
    p.scatter('x', 'y', source=source, size=4, fill_color='green', line_color=None, line_alpha=0.2, legend_label=f"{object_name}: Observational data")

    # Add the error bars using segment
    p.segment(x0='x', y0='lower', x1='x', y1='upper', source=source, color='gray', line_alpha=0.7)

    # Create the Predicted figure
    ypred = list(predicted_targets_dic.values())

    filtered_df = interpolate_df(dataset=training_dataset,
                                 predicted_targets_dic=predicted_targets_dic, print_results_=False)

    if __print_results__:
        print('------------- Interpolated spectrum based on the ML predicted targets --------------')
        print(filtered_df)

    # Add the scatter plot
    p.line(
        x=x_pred['wl'],
        y=filtered_df.drop(columns=['gravity', 'c_o_ratio', 'metallicity', 'temperature', 'is_augmented']).values[0],
        line_width=1,
        legend_label='ML Predicted:' + ', '.join([['log𝑔= ', 'C/O= ', '[M/H]= ', 'T= '][i] + str(np.round(ypred[i], 2)) for i in range(4)])
    )

    # Increase size of x and y ticks
    p.title.text_font_size = '12pt'
    p.xaxis.major_label_text_font_size = '12pt'
    p.xaxis.axis_label_text_font_size = '12pt'
    p.yaxis.major_label_text_font_size = '12pt'
    p.yaxis.axis_label_text_font_size = '12pt'

    p.legend.location = "bottom_left"
    p.legend.background_fill_color = 'white'
    p.legend.background_fill_alpha = 0.5


    show(p)

def plot_pred_vs_obs_errorbar_stat(  stat_df,
                                confidence_level,
                                object_name,
                                x_obs,
                                y_obs,
                                y_obs_err,
                                training_datasets,
                                x_pred,
                                predicted_targets_dic,
                                radius,
                                __print_results__ = False):
    """
    Plot observed spectra with error bars and predicted spectra with confidence intervals.

    Parameters
    ----------
    stat_df : DataFrame
        DataFrame containing the calculated statistics.
    confidence_level : float
        Confidence level for the confidence intervals.
    object_name : str
        Name of the object being plotted.
    x_obs : list
        List of x-axis values for the observed spectra.
    y_obs : list
        List of y-axis values for the observed spectra.
    y_obs_err : list
        List of error values corresponding to the observed spectra.
    training_datasets : optional
        Training datasets used for prediction. Default is None.
    predicted_targets_dic : optional
        Dictionary of predicted targets. Default is None.
    # bd_object_class : optional
    #     Object class. Default is None.
    __print_results__ : bool
        True or False.
    """

    chi2_stat, p_value = chi_square_test(
                            x_obs=x_obs,
                            y_obs=y_obs,
                            yerr_obs=y_obs_err,

                            x_pre=stat_df['wl'][::-1],
                            y_pre=stat_df['mean'],
                            yerr_pre=stat_df['std_values'],
                            radius=radius,
                            __plot_results__=False,
                            __print_results__=True)

    if __print_results__:
        print('*'*10+ ' Predicted Targets dic ' + '*'*10 )
        print(predicted_targets_dic)

    # Create a figure


    # Create ML figure
    p = figure(
        title=object_name+': Observational vs. ML Predicted Spectra'+' [𝛘2='+str(chi2_stat)+', p-value='+ str(p_value)+']',
        x_axis_label='Features (Wavelength [μm])',
        y_axis_label='Absolute Flux (F𝜈) [erg/s/cm2/Hz]',
        y_axis_type="log",
        width=1000,
        height=400
    )

    # Create the Observationa  figure * * * * * * * * * * * * * * * * *
    max_error_threshold = 0.8

    # Calculate adjusted error bar coordinates
    upper = np.minimum(y_obs + y_obs_err, y_obs + y_obs * max_error_threshold)
    lower = np.maximum(y_obs - y_obs_err, y_obs - y_obs * max_error_threshold)# Sample data

    # Create a ColumnDataSource to store the data
    source = ColumnDataSource(data=dict(x=x_obs, y=y_obs, upper=upper, lower=lower))

    # Add the scatter plot
    p.scatter('x', 'y', source=source, size=4, fill_color='green', line_color=None, line_alpha=0.2,
              legend_label=f"{object_name}: Observational data")

    # Add the error bars using segment
    p.segment(x0='x', y0='lower', x1='x', y1='upper', source=source, color='gray', line_alpha=0.7)


    # Create the ML Predicted figure * * * * * * * * * * * * * * * * *



    p.line(x=stat_df['wl'][::-1],
           y=stat_df['mean'],
           color='blue', line_width=2,
           legend_label='ML Predicted:' + ', '.join(
               [['log𝑔= ', 'C/O= ', '[M/H]= ', 'T= '][i] + str(np.round(list(predicted_targets_dic.values())[i], 2)) for i in range(4)])+
                ', R='+str(np.round(radius,2))+' Rjup'
          )

    # Plot the shaded regions for confidence intervals
    p.varea(
        x=stat_df['wl'][::-1],
        y1=stat_df['confidence_level_lower'],
        y2=stat_df['confidence_level_upper'],
        fill_color='red',
        fill_alpha=0.8,
        legend_label='Confidence Level: {}%'.format(confidence_level)
    )

    # Plot the shaded regions for 1 sigma
    p.varea(
        x=stat_df['wl'][::-1],
        y1=stat_df['mean'] - stat_df['std_values'],
        y2=stat_df['mean'] + stat_df['std_values'],
        fill_color='green',
        fill_alpha=0.4,
        legend_label='1σ'
    )






    # Customize the plot
    p.title.text_font_size = '12pt'
    p.xaxis.major_label_text_font_size = '12pt'
    p.xaxis.axis_label_text_font_size = '12pt'
    p.yaxis.major_label_text_font_size = '12pt'
    p.yaxis.axis_label_text_font_size = '12pt'
    p.legend.location = "bottom_left"
    p.legend.background_fill_color = 'white'
    p.legend.background_fill_alpha = 0.5
    p.legend.click_policy = 'hide'

    # Print the results if specified
    if __print_results__:
        print("Printing results:")
        print(stat_df.head(5))

    show(p)


def calculate_confidence_intervals_std_df(dataset_df,
                                          __print_results__ = False,
                                          __plot_calculate_confidence_intervals_std_df__ = False
                                         ):
    """
    Calculate confidence intervals and other statistics for a DataFrame.

    Parameters
    ----------
    dataset_df : DataFrame
        The input DataFrame containing the data.
    __print_results__ : bool
        True or False.
    __plot_calculate_confidence_intervals_std_df__ : bool
        True or False.

    Returns
    -------
    DataFrame
        A DataFrame with the calculated statistics.
    """

    # Copy the dataset to avoid modifying the original DataFrame
    df3 = dataset_df.copy()

    confidence_level = 0.95  # Confidence level (e.g., 95% confidence)

    # Calculate the sample size
    n = len(df3)

    # Calculate the mean and standard deviation for each column
    mean_values = df3.mean()
    std_values = df3.std()

    # Calculate the standard error
    se_values = std_values / np.sqrt(n)

    # Calculate the t-value for the desired confidence level and degrees of freedom
    t_value = stats.t.ppf((1 + confidence_level) / 2, df = n - 1)

    # Calculate the confidence interval for each column
    stat_df = pd.DataFrame(columns=['confidence_level_lower', 'confidence_level_upper'], index=None)
    stat_df = stat_df.astype(np.float64)

    for column in df3.columns:
        lower_bound = mean_values[column] - t_value * se_values[column]
        upper_bound = mean_values[column] + t_value * se_values[column]
        stat_df.loc[column] = [lower_bound, upper_bound]

    # Add additional columns to the DataFrame
    stat_df['mean'] = mean_values
    stat_df['std_values'] = std_values
    stat_df['wl'] = np.float64(df3.columns)

    if __print_results__:
        print(stat_df.head(5))

    if __plot_calculate_confidence_intervals_std_df__:
        # Plot the results
        x = np.round(df3.columns, 2)
        p = figure(
            title='Mean with Confidence Intervals',
            x_axis_label='Features (Wavelength [μm])',
            y_axis_label='Flux (F𝜈) [erg/s/cm2/Hz]',
            y_axis_type="log",
            width=1000,
            height=400
        )

        p.line(x = stat_df['wl'][::-1],
               y = stat_df['mean'],
               color = 'blue',
               line_width = 2,
               legend_label='Mean')#+', '.join([['log𝑔= ','C/O= ', '[M/H]= ', 'T= '][i]+str(np.round(ypred[i],2)) for i in  range(4)])))

        p.varea(
            x  = stat_df['wl'][::-1],
            y1 = stat_df['confidence_level_lower'],
            y2 = stat_df['confidence_level_upper'],
            fill_color = 'red',
            fill_alpha=0.8,
            legend_label='Confidence Level: {}%'.format(confidence_level)
        )

        p.varea(
            x  = stat_df['wl'][::-1],
            y1 = stat_df['mean'] - stat_df['std_values'],
            y2 = stat_df['mean'] + stat_df['std_values'],
            fill_color = 'green',
            fill_alpha = 0.4,
            legend_label = '1σ'
        )

        p.legend.click_policy = 'hide'

        show(p)

    return stat_df


def plot_with_errorbars(x_obs, y_obs, err_obs, x_pre, y_pre, err_pre, title="Data with Error Bars"):
    """
    Create a Bokeh plot with custom error bars for two datasets.

    Parameters
    ----------
    x_obs : array
        X-axis values for observed dataset.
    y_obs : array
        Y-axis values for observed dataset.
    err_obs : array
        Error bars for observed dataset (positive values).
    x_pre : array
        X-axis values for predicted dataset.
    y_pre : array
        Y-axis values for predicted dataset.
    err_pre : array
        Error bars for predicted dataset (positive values).
    title : str
        Title of the plot (default is "Data with Error Bars").

    Returns
    -------
    None
        (Displays the plot).
    """

    # Calculate upper and lower error bars for observed dataset
    upper_err_obs = [y_i + err for y_i, err in zip(y_obs, err_obs)]
    lower_err_obs = [y_i - err for y_i, err in zip(y_obs, err_obs)]

    # Calculate upper and lower error bars for predicted dataset
    upper_err_pre = [y_i + err for y_i, err in zip(y_pre, err_pre)]
    lower_err_pre = [y_i - err for y_i, err in zip(y_pre, err_pre)]

    # Create Bokeh ColumnDataSources for both datasets
    source_obs = ColumnDataSource(data=dict(x_obs=x_obs, y_obs=y_obs, upper_err_obs=upper_err_obs, lower_err_obs=lower_err_obs))
    source_pre = ColumnDataSource(data=dict(x_pre=x_pre, y_pre=y_pre, upper_err_pre=upper_err_pre, lower_err_pre=lower_err_pre))

    p = figure(
        x_axis_label='Features (Wavelength [𝜇m])',
        y_axis_label='Flux (F𝜈) [erg/s/cm2/Hz]',
        width=800, height=300,
        y_axis_type='log',
        title=title
    )

    # Plot data points for observed dataset
    p.circle(x='x_obs', y='y_obs', source=source_obs, size=3, color="blue", legend_label="Observed")

    # Plot custom error bars for observed dataset using the segment glyph
    p.segment(x0='x_obs', y0='lower_err_obs', x1='x_obs', y1='upper_err_obs', line_color="grey", source=source_obs)

    # Plot data points for predicted dataset
    p.square(x='x_pre', y='y_pre', source=source_pre, size=3, color="red", legend_label="Predicted")

    # Plot custom error bars for predicted dataset using the segment glyph
    p.segment(x0='x_pre', y0='lower_err_pre', x1='x_pre', y1='upper_err_pre', line_color="grey", source=source_pre)

    # Add legend
    p.legend.location = "top_left"
    p.legend.click_policy = "hide"

    # Increase size of x and y ticks
    p.title.text_font_size = '12pt'
    p.xaxis.major_label_text_font_size = '12pt'
    p.xaxis.axis_label_text_font_size = '12pt'
    p.yaxis.major_label_text_font_size = '12pt'
    p.yaxis.axis_label_text_font_size = '12pt'


    p.legend.location = "top_right"
    p.legend.background_fill_color = 'white'
    p.legend.background_fill_alpha = 0.5

    # Show the plot
    output_notebook()
    show(p)


def chi_square_test(x_obs, y_obs, yerr_obs,
                    x_pre, y_pre, yerr_pre,
                    radius,
                    __plot_results__ = False,
                    __print_results__ = True):
    """
    Perform the chi-square test to evaluate the similarity between two datasets with error bars.

    Parameters
    ----------
    x_obs : array
        The x-coordinates of the observed dataset.
    y_obs : array
        The y-coordinates of the observed dataset.
    yerr_obs : array
        The error bars associated with the observed dataset.
    x_pre : array
        The x-coordinates of the predicted dataset.
    y_pre : array
        The y-coordinates of the predicted dataset.
    yerr_pre : array
        The error bars associated with the predicted dataset.
    radius : float
        The radius value for comparison of points between datasets.
    __plot_results__ : bool, optional
        If True, plot the results of the chi-square test. Defaults to False.
    __print_results__ : bool, optional
        If True, print the results of the chi-square test. Defaults to True.

    Returns
    -------
    float
        The chi-square test statistic.
    float
        The p-value.

    Raises
    ------
    ValueError
        If the lengths of the datasets or error bars are not equal.
    """

    indices_to_remove = np.where(np.isnan(y_obs))[0]

    # Create a boolean mask with True for elements to keep, and False for elements to remove
    mask = np.ones(len(y_obs), dtype=bool)
    mask[indices_to_remove] = False
    y_obs = y_obs[mask]
    yerr_obs = yerr_obs[mask]
    x_obs = x_obs[mask]

    # Convert input to NumPy arrays for easier calculations
    data1 = np.asarray(y_obs)
    data2 = np.asarray(y_pre)
    error1 = np.asarray(yerr_obs)
    error2 = np.asarray(yerr_pre)

    num_points = len(x_pre)
    # print(y_obs)

    # Interpolate datasets if they have different lengths
    if len(data1) != len(data2):
        f1 = interp1d(x_obs, data1, kind='cubic', fill_value='extrapolate') #interp1d(x1, data1, kind='quadratic')
        f2 = interp1d(x_pre, data2, kind='cubic', fill_value='extrapolate')
        data1 = f1(x_pre)
        data2 = f2(x_pre)

        f_error1 = interp1d(x_obs, error1, kind='cubic', fill_value='extrapolate')
        f_error2 = interp1d(x_pre, error2, kind='cubic', fill_value='extrapolate')
        error1 = f_error1(x_pre)
        error2 = f_error2(x_pre)


    # Calculate the chi-square test statistic
    chi2_stat = np.round( np.sum(((data1 - data2) / np.sqrt(error1**2 + error2**2))**2), 2)
    # print(data1,data2)

    # Calculate the degrees of freedom
    degrees_of_freedom = len(data1) - 1

    # Calculate the p-value using the chi-square distribution
    p_value = "{:.2e}".format( 1.0 - chi2.cdf(chi2_stat, degrees_of_freedom) )
    # p_value = '{:.2e}'.p_value

    if __plot_results__:
        plot_with_errorbars(x_pre, data1, error1,
                            x_pre, data2, error2,
                            title=f"Radius={'{:.2f}'.format(radius)} R_Jup:  𝛘2={chi2_stat}, p-value={p_value}")

    if __print_results__:
        print( f"Radius = {'{:.2f}'.format(radius)} R_Jup:  𝛘2 = {chi2_stat}, p-value = {p_value}")


    return chi2_stat, p_value


def plot_chi_square_p_value(radius, chi_square_list, p_value_list,
                            ):
    """
    Plot two lines on the same plot with twin y-axis.

    Parameters
    ----------
    radius : array
        The x-axis values.
    chi_square_list : array
        The y-axis values for the first line.
    p_value_list : array
        The y-axis values for the second line.

    Returns
    -------
    None
        (displays the plot).
    """

    # Create a ColumnDataSource to hold the data
    source = ColumnDataSource(data=dict(radius=radius, chi_square_list=chi_square_list, p_value_list=p_value_list))

    # Create the Bokeh figure
    fig = figure(width=800, height=400, title="Chi-square and p-value", x_axis_label="Radius [R_Jup]",
                 y_axis_label="Statistic Test Metric",
                 y_axis_type="log",
                 y_range=(0.01, max(chi_square_list) + 0.20 * max(chi_square_list)))

    # Plot the first line
    fig.scatter('radius', 'chi_square_list', source=source, color='blue', marker='circle',
                legend_label='𝛘2 value')

    # Plot the second line
    fig.scatter('radius', 'p_value_list', source=source, color='red', marker='circle', legend_label='p-value')

    # Add the horizontal line at y=0.05
    fig.line([min(radius), max(radius)], [0.05, 0.05],
             line_color='black', line_dash='dashed', line_width=2, line_alpha=0.7,
             legend_label='Significance Level (𝛼)=0.05')

    # Add legend
    fig.legend.location = "bottom_left"
    fig.legend.click_policy = "hide"
    fig.legend.background_fill_color = 'white'
    fig.legend.background_fill_alpha = 0.5

    # Increase size of x and y ticks
    fig.title.text_font_size = '12pt'
    fig.xaxis.major_label_text_font_size = '12pt'
    fig.xaxis.axis_label_text_font_size = '12pt'
    fig.yaxis.major_label_text_font_size = '12pt'
    fig.yaxis.axis_label_text_font_size = '12pt'

    # Show the plot
    show(fig)



def find_closest_chi_square(df, chi_square_statistic_list):
    """
    Find the closest chi-square test and p-value for a given degrees of freedom (df).

    Parameters
    ----------
    df : int
        Degrees of freedom for the chi-square test.
    chi_square_statistic_list : list
        List of chi-square test statistics.

    Returns
    -------
    closest_chi_square : float
        The closest chi-square test statistic.
    closest_p_value : float
        The p-value corresponding to the closest chi-square test.
    """

    # Significance level (alpha)
    alpha = 0.05

    # Calculate the critical value for the given significance level and df
    critical_value = chi2.ppf(1 - alpha, df)

    closest_chi_square = None
    closest_p_value = None
    closest_difference = float('inf')

    for chi_square in chi_square_statistic_list:
        # Calculate the p-value
        p_value = 1.0 - chi2.cdf(chi_square, df)

        # Calculate the absolute difference between the chi-square statistic and the critical value
        difference = abs(chi_square - critical_value)

        # Update closest_chi_square and closest_p_value if the current test has a smaller difference
        if difference < closest_difference:
            closest_chi_square = chi_square
            closest_p_value = p_value
            closest_difference = difference

    return closest_chi_square, closest_p_value

# Example usage with df = 103 and chi_square_list containing chi-square statistics
# df_value = 103
# chi_square_list = [93, 32,  150.456789123, 120.789123456]  # Replace with actual chi-square statistics



# check_chi_square(103, chi_square_list)


def plot_scatter_x_y (x, y,
                        plot_title="Scatter Plot",
                        x_label="X-axis Label",
                        y_label="Y-axis Label",
                        plot_width = 800,
                        plot_height = 400):

    # Create a Bokeh figure
    p = figure(
        title=plot_title,
        x_axis_label=x_label,
        y_axis_label=y_label,
        width=plot_width,  # Width of the plot in pixels
        height=plot_height,  # Height of the plot in pixels
        tools="pan,box_zoom,reset,save",  # Enable interactive tools
    )

    # Add a scatter plot
    p.circle(
        x,
        y,
        size=10,  # Size of the data points
        color="blue",
        legend_label="Data",
        line_color="black",  # Color of the circle border
        line_width=2,  # Width of the circle border line
        fill_alpha=0.6,  # Transparency of the circles (0-1)
    )

    # Customize the appearance
    p.title.text_font_size = "16px"
    p.xaxis.axis_label_text_font_size = "14px"
    p.yaxis.axis_label_text_font_size = "14px"
    p.legend.label_text_font_size = "12px"

    # Show grid lines
    p.grid.visible = True
    p.grid.grid_line_color = "gray"
    p.grid.grid_line_dash = "dotted"
    p.grid.grid_line_alpha = 0.5

    # Show the plot
    show(p)


def plot_filtered_dataframe(dataset, filter_bounds, feature_to_plot, title_label, wl_synthetic, __reference_data__):
    """
    Plot a DataFrame with a single x-axis (using column names) and multiple y-axes.

    Parameters:
        - df (pd.DataFrame): DataFrame containing the data to be plotted.
    """

    filtered_df = dataset.copy()
    for feature, bounds in filter_bounds.items():
        lower_bound, upper_bound = bounds
        filtered_df = filtered_df[(filtered_df[feature] >= lower_bound) & (filtered_df[feature] <= upper_bound)]

    filtered_df2 = filtered_df.sort_values(feature_to_plot, ascending=False).iloc[::1, 4:-1][::-1]

    fig, ax = plt.subplots(figsize=(10, 3))

    x = filtered_df2.columns
    df_transposed = filtered_df2.T  # Transpose the DataFrame

    # Define a color palette
    num_colors = len(df_transposed.columns)  # Number of colors needed (excluding x-axis)
    colors = sns.color_palette('magma', num_colors)

    for i, col in enumerate(df_transposed.columns):
        # print(col)
        if col != 'x':  # Skip the x-axis column
            ax.semilogy(wl_synthetic, df_transposed[col],
                        # label=data[col][:4].values,
                        color=colors[i], alpha=0.7)

    # print(filtered_data.T[col][:4].values[0])
    ax.set_xlabel('Features (Wavelength [$\mu$m])')
    ax.set_ylabel(r'F$_{\nu}$  [erg/cm$^2$/s/Hz]')
    dict_features = {'temperature': 'Effective Temperature', 'gravity': 'Gravity', 'metallicity': 'Metallicity',
                     'c_o_ratio': 'Carbon-to-oxygen ratio'}
    ax.set_title(dict_features[feature_to_plot] + " " + title_label)
    # ax.legend()

    # Get the minimum and maximum values from the data
    # vmin = df_transposed.values.min()
    # vmax = df_transposed.values.max()

    # Add colorbar
    cmap = sns.color_palette('magma', as_cmap=True)
    cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap,
                                              norm=plt.Normalize(vmin=filter_bounds[feature_to_plot][0],
                                                                 vmax=filter_bounds[feature_to_plot][1])), ax=ax)
    # dict_features2 = {'temperature':'T [K]', 'gravity':'log$g$', 'metallicity':'[M/H]', 'c_o_ratio':'C/O ratio'}
    dict_features = {'temperature': 'T$_{eff}$ [K]', 'gravity': 'log$g$', 'metallicity': '[M/H]', 'c_o_ratio': 'C/O'}
    cbar.set_label(dict_features[feature_to_plot])

    # plt.savefig(os.path.join(__reference_data__, 'figures', feature_to_plot + "_trainin_examples.pdf"), dpi=500,
    #             bbox_inches='tight')

    plt.show()


def plot_model_loss(history=None, title=None):
    """
    Plot the trained model history for all individual target features
    """
    # from bokeh.plotting import figure, show
    # from bokeh.models import Legend

    # history = self.trained_model_history if history is None else history
    # Define the epochs as a list
    epochs = list(range(len(history['loss'])))

    # Define colorblind-friendly colors
    colors = ['#d62728', '#ff7f0e', '#2ca02c', '#9467bd', '#8c564b']

    # Create a new figure
    p = figure(title=title, width=1000, height=300, y_axis_type='log', x_axis_label='Epochs', y_axis_label='Loss')

    # Add the data lines to the figure with colorblind-friendly colors and increased line width
    p.line(epochs, history['loss'], line_color=colors[0], line_dash='solid', line_width=2,
           legend_label='Total loss')
    p.line(epochs, history['val_loss'], line_color=colors[0], line_dash='dotted', line_width=2)

    p.line(epochs, history['gravity_loss'], line_color=colors[1], line_dash='solid', line_width=2,
           legend_label='gravity')
    p.line(epochs, history['val_gravity_loss'], line_color=colors[1], line_dash='dotted', line_width=2)

    p.line(epochs, history['c_o_ratio_loss'], line_color=colors[2], line_dash='solid', line_width=2,
           legend_label='c_o_ratio')
    p.line(epochs, history['val_c_o_ratio_loss'], line_color=colors[2], line_dash='dotted', line_width=2)

    p.line(epochs, history['metallicity_loss'], line_color=colors[3], line_dash='solid', line_width=2,
           legend_label='metallicity')
    p.line(epochs, history['val_metallicity_loss'], line_color=colors[3], line_dash='dotted', line_width=2)

    p.line(epochs, history['temperature_loss'], line_color=colors[4], line_dash='solid', line_width=2,
           legend_label='temperature')
    p.line(epochs, history['val_temperature_loss'], line_color=colors[4], line_dash='dotted', line_width=2)

    # Increase size of x and y ticks
    p.title.text_font_size = '14pt'
    p.xaxis.major_label_text_font_size = '12pt'
    p.xaxis.axis_label_text_font_size = '12pt'
    p.yaxis.major_label_text_font_size = '12pt'
    p.yaxis.axis_label_text_font_size = '12pt'

    # display legend in top left corner (default is top right corner)
    p.legend.location = "bottom_left"

    # change appearance of legend text
    # p.legend.label_text_font = "times"
    # p.legend.label_text_font_style = "italic"
    # p.legend.label_text_color = "navy"

    # change border and background of legend
    # p.legend.border_line_width = 3
    # p.legend.border_line_color = "navy"
    # p.legend.border_line_alpha = 0.8
    p.legend.background_fill_color = 'white'
    p.legend.background_fill_alpha = 0.5

    # Show the plot
    show(p)



from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()



setup(
    name='TelescopeML',
    version = '0.0.2',  # MAJOR.MINOR.PATCH
    description = 'An End-to-End Python Package for Interpreting Telescope Datasets through Training Machine Learning Models, Generating Statistical Reports, and Visualizing Results',
    long_description = long_description,
    long_description_content_type='text/markdown',
    author = 'Ehsan (Sam) Gharib-Nezhad',
    author_email = 'e.gharibnezhad@gmail.com',
    url = 'https://ehsangharibnezhad.github.io/TelescopeML',
    license = 'GPL-3.0',
    download_url = 'https://github.com/EhsanGharibNezhad/TelescopeML',
    classifiers = [
                  'Intended Audience :: Science/Research',
                  'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                  'Operating System :: OS Independent' ,
                  'Programming Language :: Python',
                  'Programming Language :: Python :: 3',
                  'Topic :: Scientific/Engineering :: Astronomy',
                  'Topic :: Software Development :: Libraries :: Python Modules'
  ],
  packages=find_packages(exclude=('tests', 'docs')),
  install_requires=[
                  'numpy==1.26.1',
                  'bokeh',
                  'pandas',
                  'astropy',
                  'matplotlib',
                  'seaborn==0.12.2',
                  'sphinx==7.2.6',
                  'scipy==1.11.1',
                  'keras==2.14.0',
                  'tensorflow==2.14.0',
                  'jupyterlab',
                  'sphinx',
                  'spectres==2.2.0',
                  'scikit-learn==1.3.0',
  ],
    zip_safe = False,
)



# try:
#     from pkg_resources import declare_namespace
#     declare_namespace(__name__)
# except ImportError:
#     from pkgutil import extend_path
#     __path__ = extend_path(__path__, __name__)

from os.path import dirname, basename, isfile, join
import glob
modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [ basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]

# try:
#     from pkg_resources import declare_namespace
#     declare_namespace(__name__)
# except ImportError:
#     from pkgutil import extend_path
#     __path__ = extend_path(__path__, __name__)

__version__ = "0.0.1"